from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from typing import (Any, Protocol, TypeVar)
from ..arctrl_core.Helper.collections_ import ResizeArray_map
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..dynamic_obj.hash_codes import merge_hashes
from ..dynamic_obj.property_helper import PropertyHelper
from ..fable_library.map_util import (get_item_from_dict, try_get_value, add_to_dict)
from ..fable_library.option import (default_arg, value as value_3, some, map)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (iterate, filter as filter_2, choose, exists, to_list, delay, enumerate_while, singleton as singleton_2, empty, try_find, map as map_1, enumerate_from_functions, append, collect)
from ..fable_library.seq2 import List_distinctBy
from ..fable_library.string_ import starts_with_exact
from ..fable_library.types import (Array, FSharpRef)
from ..fable_library.util import (is_iterable, IEnumerable, equals, identity_hash, string_hash, IEnumerable_1, get_enumerator, IEnumerator, curry2, ignore, structural_hash, is_array_like)
from .ldcontext import LDContext

__G = TypeVar("__G")

__F = TypeVar("__F")

__D = TypeVar("__D")

__C = TypeVar("__C")

__B = TypeVar("__B")

__A = TypeVar("__A")

def ActivePattern__007CNonStringEnumerable_007C__007C(o: Any=None) -> IEnumerable[Any] | None:
    if str(type(o)) == "<class \'str\'>":
        return None

    elif is_iterable(o):
        return o

    else: 
        return None



def DynamicObj_DynamicObj__DynamicObj_HasProperty_Z721C83C5(this: DynamicObj, property_name: str) -> bool:
    return this.TryGetPropertyValue(property_name) is not None


class ILDObject(Protocol):
    @property
    @abstractmethod
    def AdditionalType(self) -> Array[str]:
        ...

    @AdditionalType.setter
    @abstractmethod
    def AdditionalType(self, __arg0: Array[str]) -> None:
        ...

    @property
    @abstractmethod
    def Id(self) -> str:
        ...

    @property
    @abstractmethod
    def SchemaType(self) -> Array[str]:
        ...

    @SchemaType.setter
    @abstractmethod
    def SchemaType(self, __arg0: Array[str]) -> None:
        ...


def _expr1192() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDValue", None, LDValue)


class LDValue:
    def __init__(self, value: Any=None, value_type: str | None=None) -> None:
        self.value_type_004054: str = default_arg(value_type, "string")
        self.value_004055: Any = value

    @property
    def Value(self, __unit: None=None) -> Any:
        this: LDValue = self
        return this.value_004055

    @Value.setter
    def Value(self, v: Any=None) -> None:
        this: LDValue = self
        this.value_004055 = v

    @property
    def ValueType(self, __unit: None=None) -> str:
        this: LDValue = self
        return this.value_type_004054

    @ValueType.setter
    def ValueType(self, v: str) -> None:
        this: LDValue = self
        this.value_type_004054 = v

    def __eq__(self, other: Any=None) -> bool:
        this: LDValue = self
        return equals(this.Value, other.Value) if isinstance(other, LDValue) else False

    def __hash__(self, __unit: None=None) -> int:
        this: LDValue = self
        return merge_hashes(123, identity_hash(this.Value))


LDValue_reflection = _expr1192

def LDValue__ctor_77809003(value: Any=None, value_type: str | None=None) -> LDValue:
    return LDValue(value, value_type)


def _expr1193() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDRef", None, LDRef)


class LDRef:
    def __init__(self, id: str) -> None:
        self.id_004075: str = id

    @property
    def Id(self, __unit: None=None) -> str:
        this: LDRef = self
        return this.id_004075

    @Id.setter
    def Id(self, v: str) -> None:
        this: LDRef = self
        this.id_004075 = v

    def __eq__(self, other: Any=None) -> bool:
        this: LDRef = self
        return (this.Id == other.Id) if isinstance(other, LDRef) else False

    def __hash__(self, __unit: None=None) -> int:
        this: LDRef = self
        return merge_hashes(123, string_hash(this.Id))


LDRef_reflection = _expr1193

def LDRef__ctor_Z721C83C5(id: str) -> LDRef:
    return LDRef(id)


def _expr1198() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDGraph", None, LDGraph, DynamicObj_reflection())


class LDGraph(DynamicObj):
    def __init__(self, id: str | None=None, nodes: Array[LDNode] | None=None, context: LDContext | None=None) -> None:
        super().__init__()
        this: FSharpRef[LDGraph] = FSharpRef(None)
        this.contents = self
        self.id_004094: str | None = id
        self.mappings: Any = dict([])
        self.init_004090: int = 1
        if context is None:
            pass

        else: 
            ctx: LDContext = context
            this.contents.SetContext(ctx)

        if nodes is None:
            pass

        else: 
            def action(node: LDNode) -> None:
                add_to_dict(self.mappings, node.Id, node)

            iterate(action, nodes)


    @property
    def Id(self, __unit: None=None) -> str | None:
        this: LDGraph = self
        return this.id_004094

    @Id.setter
    def Id(self, v: str | None=None) -> None:
        this: LDGraph = self
        this.id_004094 = v

    @property
    def Nodes(self, __unit: None=None) -> Array[LDNode]:
        this: LDGraph = self
        return list(this.mappings.values())

    def ContainsNode(self, id: str) -> bool:
        this: LDGraph = self
        return this.mappings.has(id)

    def GetNode(self, id: str) -> LDNode:
        this: LDGraph = self
        return get_item_from_dict(this.mappings, id)

    def TryGetNode(self, id: str) -> LDNode | None:
        this: LDGraph = self
        match_value: tuple[bool, LDNode]
        out_arg: LDNode = None
        def _arrow1194(__unit: None=None) -> LDNode:
            return out_arg

        def _arrow1195(v: LDNode) -> None:
            nonlocal out_arg
            out_arg = v

        match_value = (try_get_value(this.mappings, id, FSharpRef(_arrow1194, _arrow1195)), out_arg)
        return match_value[1] if match_value[0] else None

    def AddNode(self, node: LDNode) -> None:
        this: LDGraph = self
        id: str = node.Id
        match_value: LDNode | None = this.TryGetNode(id)
        if match_value is None:
            add_to_dict(this.mappings, id, node)

        else: 
            existing_node: LDNode = match_value
            node.MergeAppendInto_InPlace(existing_node, this)


    def Compact_InPlace(self, context: LDContext | None=None) -> None:
        this: LDGraph = self
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        def action(node: LDNode) -> None:
            node.Compact_InPlace(context_1)

        iterate(action, this.Nodes)

    def SetContext(self, context: LDContext) -> None:
        this: LDGraph = self
        this.SetProperty("@context", context)

    @staticmethod
    def set_context(context: LDContext) -> Callable[[__G], None]:
        def _arrow1196(roc: __G | None=None) -> None:
            roc.SetContext(context)

        return _arrow1196

    def TryGetContext(self, __unit: None=None) -> LDContext | None:
        this: LDGraph = self
        match_value: Any | None = this.TryGetPropertyValue("@context")
        if match_value is not None:
            o: Any = value_3(match_value)
            return o if isinstance(o, LDContext) else None

        else: 
            return None


    @staticmethod
    def try_get_context(__unit: None=None) -> Callable[[__F], LDContext | None]:
        def _arrow1197(roc: __F | None=None) -> LDContext | None:
            return roc.TryGetContext()

        return _arrow1197

    def RemoveContext(self, __unit: None=None) -> bool:
        this: LDGraph = self
        return this.RemoveProperty("@context")

    def GetDynamicPropertyHelpers(self, __unit: None=None) -> IEnumerable_1[PropertyHelper]:
        this: LDGraph = self
        def predicate(ph: PropertyHelper) -> bool:
            return not (True if starts_with_exact(ph.Name, "init@") else (ph.Name == "mappings"))

        return filter_2(predicate, this.GetPropertyHelpers(False))

    def GetDynamicPropertyNames(self, context: LDContext | None=None) -> IEnumerable_1[str]:
        this: LDGraph = self
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        def chooser(ph: PropertyHelper) -> str | None:
            name: str
            if context_1 is None:
                name = ph.Name

            else: 
                ctx: LDContext = context_1
                match_value: str | None = ctx.TryResolveTerm(ph.Name)
                name = ph.Name if (match_value is None) else match_value

            if name == "@context":
                return None

            else: 
                return name


        return choose(chooser, this.GetDynamicPropertyHelpers())


LDGraph_reflection = _expr1198

def LDGraph__ctor_5FC797DE(id: str | None=None, nodes: Array[LDNode] | None=None, context: LDContext | None=None) -> LDGraph:
    return LDGraph(id, nodes, context)


def _expr1285() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDNode", None, LDNode, DynamicObj_reflection())


class LDNode(DynamicObj):
    def __init__(self, id: str, schema_type: Array[str], additional_type: Array[str] | None=None, context: LDContext | None=None) -> None:
        super().__init__()
        this: FSharpRef[LDNode] = FSharpRef(None)
        self.id: str = id
        this.contents = self
        self.schema_type_0040178: Array[str] = schema_type
        self.additional_type_0040179: Array[str] = default_arg(additional_type, [])
        self.init_0040175_002D1: int = 1
        if context is None:
            pass

        else: 
            ctx: LDContext = context
            this.contents.SetContext(ctx)


    @property
    def Id(self, __unit: None=None) -> str:
        this: LDNode = self
        return this.id

    @property
    def SchemaType(self, __unit: None=None) -> Array[str]:
        this: LDNode = self
        return this.schema_type_0040178

    @SchemaType.setter
    def SchemaType(self, value: Array[str]) -> None:
        this: LDNode = self
        this.schema_type_0040178 = value

    @property
    def AdditionalType(self, __unit: None=None) -> Array[str]:
        this: LDNode = self
        return this.additional_type_0040179

    @AdditionalType.setter
    def AdditionalType(self, value: Array[str]) -> None:
        this: LDNode = self
        this.additional_type_0040179 = value

    def HasType(self, schema_type: str, context: LDContext | None=None) -> bool:
        this: LDNode = self
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        def predicate(st: str) -> bool:
            if st == schema_type:
                return True

            elif context_1 is None:
                return False

            else: 
                ctx: LDContext = context_1
                match_value: str | None = ctx.TryResolveTerm(st)
                match_value_1: str | None = ctx.TryResolveTerm(schema_type)
                if match_value is None:
                    if match_value_1 is not None:
                        schema_type_2: str = match_value_1
                        return st == schema_type_2

                    else: 
                        return False


                elif match_value_1 is None:
                    st_2: str = match_value
                    return st_2 == schema_type

                else: 
                    schema_type_1: str = match_value_1
                    st_1: str = match_value
                    return st_1 == schema_type_1



        return exists(predicate, this.SchemaType)

    def TryGetProperty(self, property_name: str, context: LDContext | None=None) -> Any | None:
        this: LDNode = self
        match_value: Any | None = this.TryGetPropertyValue(property_name)
        if match_value is None:
            match_value_1: LDContext | None
            specific_context: LDContext | None = this.TryGetContext()
            match_value_1 = LDContext.try_combine_optional(context, specific_context)
            if match_value_1 is None:
                return None

            else: 
                ctx: LDContext = match_value_1
                match_value_2: str | None = ctx.TryResolveTerm(property_name)
                if match_value_2 is None:
                    match_value_3: str | None = ctx.TryGetTerm(property_name)
                    if match_value_3 is None:
                        return None

                    else: 
                        term_1: str = match_value_3
                        return this.TryGetPropertyValue(term_1)


                else: 
                    term: str = match_value_2
                    return this.TryGetPropertyValue(term)



        else: 
            return some(value_3(match_value))


    def TryGetPropertyAsSingleton(self, property_name: str, context: LDContext | None=None) -> Any | None:
        this: LDNode = self
        match_value: Any | None = this.TryGetProperty(property_name, context)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                s: str = value_3(match_value)
                return some(s)

            elif is_iterable(value_3(match_value)):
                e: IEnumerable[Any] = value_3(match_value)
                en: IEnumerator[Any] = get_enumerator(e)
                return some(en.System_Collections_IEnumerator_get_Current()) if en.System_Collections_IEnumerator_MoveNext() else None

            else: 
                o: Any = value_3(match_value)
                return some(o)


        else: 
            return None


    def GetPropertyValues(self, property_name: str, filter: Callable[[Any, LDContext | None], bool] | None=None, context: LDContext | None=None) -> Array[Any]:
        this: LDNode = self
        def _arrow1201(_arg: Any=None) -> Callable[[LDContext | None], bool]:
            def _arrow1200(_arg_1: LDContext | None=None) -> bool:
                return True

            return _arrow1200

        filter_1: Callable[[Any, LDContext | None], bool] = default_arg(map(curry2, filter), _arrow1201)
        match_value: Any | None = this.TryGetProperty(property_name, context)
        (pattern_matching_result, s, e, o_1) = (None, None, None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                s = value_3(match_value)

            elif is_iterable(value_3(match_value)):
                pattern_matching_result = 1
                e = value_3(match_value)

            elif filter_1(value_3(match_value))(context):
                pattern_matching_result = 2
                o_1 = value_3(match_value)

            else: 
                pattern_matching_result = 3


        else: 
            pattern_matching_result = 3

        if pattern_matching_result == 0:
            return [s] if filter_1(s)(context) else []

        elif pattern_matching_result == 1:
            en: IEnumerator[Any] = get_enumerator(e)
            def _arrow1204(__unit: None=None) -> IEnumerable_1[Any]:
                def _arrow1202(__unit: None=None) -> bool:
                    return en.System_Collections_IEnumerator_MoveNext()

                def _arrow1203(__unit: None=None) -> IEnumerable_1[Any]:
                    return singleton_2(en.System_Collections_IEnumerator_get_Current()) if filter_1(en.System_Collections_IEnumerator_get_Current())(context) else empty()

                return enumerate_while(_arrow1202, delay(_arrow1203))

            return list(to_list(delay(_arrow1204)))

        elif pattern_matching_result == 2:
            return [o_1]

        elif pattern_matching_result == 3:
            return []


    def TryGetPropertyAsSingleNode(self, property_name: str, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        this: LDNode = self
        match_value: Any | None = this.TryGetPropertyAsSingleton(property_name, context)
        (pattern_matching_result, n, r_1) = (None, None, None)
        if match_value is not None:
            if isinstance(value_3(match_value), LDNode):
                pattern_matching_result = 0
                n = value_3(match_value)

            elif isinstance(value_3(match_value), LDRef):
                def _arrow1207(__unit: None=None) -> bool:
                    r: LDRef = value_3(match_value)
                    return graph is not None

                if _arrow1207():
                    pattern_matching_result = 1
                    r_1 = value_3(match_value)

                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2


        else: 
            pattern_matching_result = 2

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            match_value_1: LDNode | None = value_3(graph).TryGetNode(r_1.Id)
            return None if (match_value_1 is None) else match_value_1

        elif pattern_matching_result == 2:
            return None


    def GetPropertyNodes(self, property_name: str, filter: Callable[[LDNode, LDContext | None], bool] | None=None, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        this: LDNode = self
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        def chooser(o: Any=None) -> LDNode | None:
            (pattern_matching_result,) = (None,)
            if isinstance(o, LDRef):
                if graph is not None:
                    pattern_matching_result = 0

                else: 
                    pattern_matching_result = 2


            elif isinstance(o, LDNode):
                pattern_matching_result = 1

            else: 
                pattern_matching_result = 2

            if pattern_matching_result == 0:
                match_value: LDNode | None = value_3(graph).TryGetNode(o.Id)
                if match_value is None:
                    return None

                else: 
                    n: LDNode = match_value
                    if filter is None:
                        return n

                    elif filter(n, context_1):
                        def _arrow1216(f: Callable[[LDNode, LDContext | None], bool], o: Any=o) -> Callable[[LDNode, LDContext | None], bool]:
                            return curry2(f)

                        f_1: Callable[[LDNode, LDContext | None], bool] = map(_arrow1216, filter)
                        return n

                    else: 
                        return None



            elif pattern_matching_result == 1:
                n_1: LDNode = o
                if filter is None:
                    return n_1

                elif filter(n_1, context_1):
                    def _arrow1218(f: Callable[[LDNode, LDContext | None], bool], o: Any=o) -> Callable[[LDNode, LDContext | None], bool]:
                        return curry2(f)

                    f_3: Callable[[LDNode, LDContext | None], bool] = map(_arrow1218, filter)
                    return n_1

                else: 
                    return None


            elif pattern_matching_result == 2:
                return None


        return list(choose(chooser, this.GetPropertyValues(property_name, None, context_1)))

    def GetDynamicPropertyHelpers(self, __unit: None=None) -> IEnumerable_1[PropertyHelper]:
        this: LDNode = self
        def predicate(ph: PropertyHelper) -> bool:
            return not (True if starts_with_exact(ph.Name, "init_") else (ph.Name == "id"))

        return filter_2(predicate, this.GetPropertyHelpers(False))

    def GetPropertyNames(self, context: LDContext | None=None) -> IEnumerable_1[str]:
        this: LDNode = self
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        def chooser(ph: PropertyHelper) -> str | None:
            name: str
            if context_1 is None:
                name = ph.Name

            else: 
                ctx: LDContext = context_1
                match_value: str | None = ctx.TryResolveTerm(ph.Name)
                name = ph.Name if (match_value is None) else match_value

            if name == "@context":
                return None

            else: 
                return name


        return choose(chooser, this.GetDynamicPropertyHelpers())

    def SetProperty(self, property_name: str, value: Any=None, context: LDContext | None=None) -> Any:
        this: LDNode = self
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        def predicate(pn: str) -> bool:
            if context_1 is None:
                return pn == property_name

            else: 
                c: LDContext = context_1
                return c.PropertyNamesMatch(pn, property_name)


        property_name_1: str = default_arg(try_find(predicate, this.GetPropertyNames()), property_name)
        super().SetProperty(property_name_1,value)

    def SetOptionalProperty(self, property_name: str, value: Any | None=None, context: LDContext | None=None) -> None:
        this: LDNode = self
        if value is None:
            pass

        else: 
            v: __D = value_3(value)
            this.SetProperty(property_name, v, context)


    def HasProperty(self, property_name: str, context: LDContext | None=None) -> bool:
        this: LDNode = self
        v: Any | None = this.TryGetProperty(property_name, context)
        def _arrow1232(__unit: None=None) -> bool:
            v_2: Any = value_3(v)
            return False

        def _arrow1236(__unit: None=None) -> bool:
            e: IEnumerable[Any] = value_3(v)
            return get_enumerator(e).System_Collections_IEnumerator_MoveNext()

        return (_arrow1232() if equals(value_3(v), None) else (_arrow1236() if is_iterable(value_3(v)) else True)) if (v is not None) else False

    def SetContext(self, context: LDContext) -> None:
        this: LDNode = self
        this.SetProperty("@context", context)

    @staticmethod
    def set_context(context: LDContext) -> Callable[[__C], None]:
        def _arrow1237(roc: __C | None=None) -> None:
            roc.SetContext(context)

        return _arrow1237

    def TryGetContext(self, __unit: None=None) -> LDContext | None:
        this: LDNode = self
        match_value: Any | None = this.TryGetPropertyValue("@context")
        if match_value is not None:
            o: Any = value_3(match_value)
            return o if isinstance(o, LDContext) else None

        else: 
            return None


    @staticmethod
    def try_get_context(__unit: None=None) -> Callable[[__B], LDContext | None]:
        def _arrow1238(roc: __B | None=None) -> LDContext | None:
            return roc.TryGetContext()

        return _arrow1238

    def RemoveContext(self, __unit: None=None) -> bool:
        this: LDNode = self
        return this.RemoveProperty("@context")

    def MergeAppendInto_InPlace(self, other: LDNode, flatten_to: LDGraph | None=None) -> None:
        this: LDNode = self
        flatten_to_singleton: Callable[[Any], Any]
        if flatten_to is None:
            def _arrow1239(x: Any=None) -> Any:
                return x

            flatten_to_singleton = _arrow1239

        else: 
            graph: LDGraph = flatten_to
            def f(o: Any=None) -> Any:
                if isinstance(o, LDNode):
                    ignore(o.Flatten(graph))
                    return LDRef(o.Id)

                else: 
                    return o


            flatten_to_singleton = f

        flatten_to_ra: Callable[[Array[Any]], Array[Any]]
        if flatten_to is None:
            def _arrow1240(x_1: Array[Any]) -> Array[Any]:
                return x_1

            flatten_to_ra = _arrow1240

        else: 
            graph_1: LDGraph = flatten_to
            def _arrow1241(a: Array[Any]) -> Array[Any]:
                return ResizeArray_map(flatten_to_singleton, a)

            flatten_to_ra = _arrow1241

        flatten_to_any: Callable[[Any], Any]
        if flatten_to is None:
            def _arrow1242(x_2: Any=None) -> Any:
                return x_2

            flatten_to_any = _arrow1242

        else: 
            graph_2: LDGraph = flatten_to
            def f_2(o_1: Any=None) -> Any:
                if isinstance(o_1, LDNode):
                    ignore(o_1.Flatten(graph_2))
                    return LDRef(o_1.Id)

                else: 
                    active_pattern_result: IEnumerable[Any] | None = ActivePattern__007CNonStringEnumerable_007C__007C(o_1)
                    if active_pattern_result is not None:
                        e: IEnumerable[Any] = active_pattern_result
                        def _arrow1247(__unit: None=None, o_1: Any=o_1) -> IEnumerable_1[Any]:
                            def _arrow1244(__unit: None=None) -> IEnumerator[Any]:
                                return get_enumerator(e)

                            def _arrow1245(enumerator: IEnumerator[Any]) -> bool:
                                return enumerator.System_Collections_IEnumerator_MoveNext()

                            def _arrow1246(enumerator_1: IEnumerator[Any]) -> Any:
                                return enumerator_1.System_Collections_IEnumerator_get_Current()

                            return map_1(f_2, enumerate_from_functions(_arrow1244, _arrow1245, _arrow1246))

                        return list(to_list(delay(_arrow1247)))

                    else: 
                        return o_1



            flatten_to_any = f_2

        def to_equalitor(o_2: Any=None) -> Any:
            if isinstance(o_2, LDNode):
                return o_2.Id

            elif isinstance(o_2, LDRef):
                return o_2.Id

            else: 
                return o_2


        def action(pn: str) -> None:
            match_value: Any | None = other.TryGetProperty(pn)
            if match_value is None:
                v_3: Any = flatten_to_any(value_3(this.TryGetProperty(pn)))
                other.SetProperty(pn, v_3)

            else: 
                other_val: Any = value_3(match_value)
                this_val: Any = value_3(this.TryGetProperty(pn))
                (pattern_matching_result, e1, e2) = (None, None, None)
                active_pattern_result_1: IEnumerable[Any] | None = ActivePattern__007CNonStringEnumerable_007C__007C(this_val)
                if active_pattern_result_1 is not None:
                    active_pattern_result_2: IEnumerable[Any] | None = ActivePattern__007CNonStringEnumerable_007C__007C(other_val)
                    if active_pattern_result_2 is not None:
                        pattern_matching_result = 0
                        e1 = active_pattern_result_1
                        e2 = active_pattern_result_2

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1

                if pattern_matching_result == 0:
                    def _arrow1258(__unit: None=None, pn: Any=pn) -> IEnumerable_1[Any]:
                        def _arrow1249(v_1: Any=None) -> Any:
                            return v_1

                        def _arrow1250(__unit: None=None) -> IEnumerator[Any]:
                            return get_enumerator(e2)

                        def _arrow1251(enumerator_2: IEnumerator[Any]) -> bool:
                            return enumerator_2.System_Collections_IEnumerator_MoveNext()

                        def _arrow1252(enumerator_3: IEnumerator[Any]) -> Any:
                            return enumerator_3.System_Collections_IEnumerator_get_Current()

                        def _arrow1257(__unit: None=None) -> IEnumerable_1[Any]:
                            def _arrow1253(v_2: Any=None) -> Any:
                                return v_2

                            def _arrow1254(__unit: None=None) -> IEnumerator[Any]:
                                return get_enumerator(e1)

                            def _arrow1255(enumerator_4: IEnumerator[Any]) -> bool:
                                return enumerator_4.System_Collections_IEnumerator_MoveNext()

                            def _arrow1256(enumerator_5: IEnumerator[Any]) -> Any:
                                return enumerator_5.System_Collections_IEnumerator_get_Current()

                            return map_1(_arrow1253, enumerate_from_functions(_arrow1254, _arrow1255, _arrow1256))

                        return append(map_1(_arrow1249, enumerate_from_functions(_arrow1250, _arrow1251, _arrow1252)), delay(_arrow1257))

                    class ObjectExpr1259:
                        @property
                        def Equals(self) -> Callable[[Any, Any], bool]:
                            return equals

                        @property
                        def GetHashCode(self) -> Callable[[Any], int]:
                            return structural_hash

                    l: Array[Any] = flatten_to_ra(list(List_distinctBy(to_equalitor, to_list(delay(_arrow1258)), ObjectExpr1259())))
                    other.SetProperty(pn, l)

                elif pattern_matching_result == 1:
                    active_pattern_result_3: IEnumerable[Any] | None = ActivePattern__007CNonStringEnumerable_007C__007C(this_val)
                    if active_pattern_result_3 is not None:
                        these_vals: IEnumerable[Any] = active_pattern_result_3
                        is_contained: bool = False
                        def _arrow1264(__unit: None=None, pn: Any=pn) -> IEnumerable_1[Any]:
                            def _arrow1260(this_val_1: Any=None) -> IEnumerable_1[Any]:
                                nonlocal is_contained
                                if equals(to_equalitor(this_val_1), to_equalitor(other_val)):
                                    is_contained = True
                                    return singleton_2(flatten_to_singleton(this_val_1))

                                else: 
                                    return singleton_2(this_val_1)


                            def _arrow1261(__unit: None=None) -> IEnumerator[Any]:
                                return get_enumerator(these_vals)

                            def _arrow1262(enumerator_6: IEnumerator[Any]) -> bool:
                                return enumerator_6.System_Collections_IEnumerator_MoveNext()

                            def _arrow1263(enumerator_7: IEnumerator[Any]) -> Any:
                                return enumerator_7.System_Collections_IEnumerator_get_Current()

                            return collect(_arrow1260, enumerate_from_functions(_arrow1261, _arrow1262, _arrow1263))

                        l_1: Array[Any] = list(to_list(delay(_arrow1264)))
                        if not is_contained:
                            (l_1.append(other_val))
                            other.SetProperty(pn, l_1)


                    else: 
                        active_pattern_result_4: IEnumerable[Any] | None = ActivePattern__007CNonStringEnumerable_007C__007C(other_val)
                        if active_pattern_result_4 is not None:
                            other_vals: IEnumerable[Any] = active_pattern_result_4
                            is_contained_1: bool = False
                            def _arrow1271(__unit: None=None, pn: Any=pn) -> IEnumerable_1[Any]:
                                def _arrow1267(other_val_2: Any=None) -> IEnumerable_1[Any]:
                                    def _expr1265():
                                        nonlocal is_contained_1
                                        is_contained_1 = True
                                        return empty()

                                    def _arrow1266(__unit: None=None) -> IEnumerable_1[Any]:
                                        return singleton_2(other_val_2)

                                    return append(_expr1265() if equals(to_equalitor(this_val), to_equalitor(other_val_2)) else empty(), delay(_arrow1266))

                                def _arrow1268(__unit: None=None) -> IEnumerator[Any]:
                                    return get_enumerator(other_vals)

                                def _arrow1269(enumerator_8: IEnumerator[Any]) -> bool:
                                    return enumerator_8.System_Collections_IEnumerator_MoveNext()

                                def _arrow1270(enumerator_9: IEnumerator[Any]) -> Any:
                                    return enumerator_9.System_Collections_IEnumerator_get_Current()

                                return collect(_arrow1267, enumerate_from_functions(_arrow1268, _arrow1269, _arrow1270))

                            l_2: Array[Any] = list(to_list(delay(_arrow1271)))
                            if not is_contained_1:
                                (l_2.append(flatten_to_singleton(this_val)))
                                other.SetProperty(pn, l_2)


                        elif equals(to_equalitor(this_val), to_equalitor(other_val)):
                            pass

                        else: 
                            l_3: Array[Any] = [flatten_to_singleton(this_val), other_val]
                            other.SetProperty(pn, l_3)





        iterate(action, this.GetPropertyNames())

    def Compact_InPlace(self, context: LDContext | None=None, set_context: bool | None=None) -> None:
        this: LDNode = self
        set_context_1: bool = default_arg(set_context, False)
        context_1: LDContext | None
        specific_context: LDContext | None = this.TryGetContext()
        context_1 = LDContext.try_combine_optional(context, specific_context)
        if context_1 is not None:
            context_2: LDContext = value_3(context_1)
            if set_context_1:
                this.SetContext(context_2)

            def _arrow1273(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow1272(st: str) -> IEnumerable_1[str]:
                    match_value: str | None = context_2.TryGetTerm(st)
                    return singleton_2(st) if (match_value is None) else singleton_2(match_value)

                return collect(_arrow1272, this.SchemaType)

            new_types: Array[str] = list(to_list(delay(_arrow1273)))
            this.SchemaType = new_types

        def compact_value_in_place(o: Any=None) -> Any:
            (pattern_matching_result, n, s, e, v, x) = (None, None, None, None, None, None)
            if isinstance(o, LDNode):
                pattern_matching_result = 0
                n = o

            elif str(type(o)) == "<class \'str\'>":
                pattern_matching_result = 1
                s = o

            elif isinstance(o, LDValue):
                if is_iterable(o):
                    pattern_matching_result = 2
                    e = o

                else: 
                    pattern_matching_result = 3
                    v = o


            elif is_iterable(o):
                pattern_matching_result = 2
                e = o

            else: 
                pattern_matching_result = 4
                x = o

            if pattern_matching_result == 0:
                n.Compact_InPlace(context_1)
                return n

            elif pattern_matching_result == 1:
                return s

            elif pattern_matching_result == 2:
                en: IEnumerator[Any] = get_enumerator(e)
                def _arrow1276(__unit: None=None, o: Any=o) -> IEnumerable_1[Any]:
                    def _arrow1274(__unit: None=None) -> bool:
                        return en.System_Collections_IEnumerator_MoveNext()

                    def _arrow1275(__unit: None=None) -> IEnumerable_1[Any]:
                        return singleton_2(compact_value_in_place(en.System_Collections_IEnumerator_get_Current()))

                    return enumerate_while(_arrow1274, delay(_arrow1275))

                l: Array[Any] = list(to_list(delay(_arrow1276)))
                if len(l) == 1:
                    return l[0]

                else: 
                    return l


            elif pattern_matching_result == 3:
                return v.Value

            elif pattern_matching_result == 4:
                return x


        def action(ph: PropertyHelper) -> None:
            new_key: str | None
            if context_1 is None:
                new_key = None

            else: 
                ctx: LDContext = context_1
                match_value_1: str | None = ctx.TryGetTerm(ph.Name)
                new_key = None if (match_value_1 is None) else match_value_1

            new_value: Any = compact_value_in_place(ph.GetValue(this))
            (pattern_matching_result_1, key_1) = (None, None)
            if new_key is not None:
                if new_key != ph.Name:
                    pattern_matching_result_1 = 0
                    key_1 = new_key

                else: 
                    pattern_matching_result_1 = 1


            else: 
                pattern_matching_result_1 = 1

            if pattern_matching_result_1 == 0:
                ignore(this.RemoveProperty(ph.Name))
                this.SetProperty(key_1, new_value)

            elif pattern_matching_result_1 == 1:
                ph.SetValue(this, new_value)


        iterate(action, this.GetDynamicPropertyHelpers())

    def Flatten(self, graph: LDGraph | None=None) -> LDGraph:
        this: LDNode = self
        graph_1: LDGraph = LDGraph(None, None, this.TryGetContext()) if (graph is None) else graph
        def flatten_value(o: Any=None) -> Any:
            if isinstance(o, LDNode):
                ignore(o.Flatten(graph_1))
                return LDRef(o.Id)

            elif str(type(o)) == "<class \'str\'>":
                return o

            elif is_iterable(o):
                en: IEnumerator[Any] = get_enumerator(o)
                def _arrow1279(__unit: None=None, o: Any=o) -> IEnumerable_1[Any]:
                    def _arrow1277(__unit: None=None) -> bool:
                        return en.System_Collections_IEnumerator_MoveNext()

                    def _arrow1278(__unit: None=None) -> IEnumerable_1[Any]:
                        return singleton_2(flatten_value(en.System_Collections_IEnumerator_get_Current()))

                    return enumerate_while(_arrow1277, delay(_arrow1278))

                return list(to_list(delay(_arrow1279)))

            else: 
                return o


        def action(ph: PropertyHelper) -> None:
            ph.SetValue(this, flatten_value(ph.GetValue(this)))

        iterate(action, this.GetDynamicPropertyHelpers())
        graph_1.AddNode(this)
        return graph_1

    def Unflatten(self, graph: LDGraph) -> None:
        this: LDNode = self
        def unflatten_value(o: Any=None) -> Any:
            if isinstance(o, LDRef):
                match_value: LDNode | None = graph.TryGetNode(o.Id)
                if match_value is None:
                    return o

                else: 
                    return match_value


            elif isinstance(o, LDNode):
                o.Unflatten(graph)
                return o

            elif is_iterable(o):
                en: IEnumerator[Any] = get_enumerator(o)
                def _arrow1282(__unit: None=None, o: Any=o) -> IEnumerable_1[Any]:
                    def _arrow1280(__unit: None=None) -> bool:
                        return en.System_Collections_IEnumerator_MoveNext()

                    def _arrow1281(__unit: None=None) -> IEnumerable_1[Any]:
                        return singleton_2(unflatten_value(en.System_Collections_IEnumerator_get_Current()))

                    return enumerate_while(_arrow1280, delay(_arrow1281))

                return list(to_list(delay(_arrow1282)))

            else: 
                return o


        def action(ph: PropertyHelper) -> None:
            ph.SetValue(this, unflatten_value(ph.GetValue(this)))

        iterate(action, this.GetDynamicPropertyHelpers())

    @staticmethod
    def remove_context(__unit: None=None) -> Callable[[__A], bool]:
        def _arrow1283(roc: __A | None=None) -> bool:
            return roc.RemoveContext()

        return _arrow1283

    @staticmethod
    def try_from_dynamic_obj(dyn_obj: DynamicObj) -> LDNode | None:
        original_id: str | None
        match_value: Any | None = dyn_obj.TryGetPropertyValue("@id")
        if match_value is not None:
            o: Any = value_3(match_value)
            original_id = o if (str(type(o)) == "<class \'str\'>") else None

        else: 
            original_id = None

        original_type: Array[str] | None
        match_value_1: Any | None = dyn_obj.TryGetPropertyValue("@type")
        (pattern_matching_result, ra, singleton) = (None, None, None)
        if match_value_1 is not None:
            if is_array_like(value_3(match_value_1)):
                pattern_matching_result = 0
                ra = value_3(match_value_1)

            elif str(type(value_3(match_value_1))) == "<class \'str\'>":
                pattern_matching_result = 1
                singleton = value_3(match_value_1)

            else: 
                pattern_matching_result = 2


        else: 
            pattern_matching_result = 2

        if pattern_matching_result == 0:
            original_type = ra

        elif pattern_matching_result == 1:
            original_type = [singleton]

        elif pattern_matching_result == 2:
            original_type = None

        (pattern_matching_result_1, id, st) = (None, None, None)
        if original_id is not None:
            if original_type is not None:
                pattern_matching_result_1 = 0
                id = original_id
                st = original_type

            else: 
                pattern_matching_result_1 = 1


        else: 
            pattern_matching_result_1 = 1

        if pattern_matching_result_1 == 0:
            def _arrow1284(__unit: None=None) -> Array[str] | None:
                match_value_3: Any | None = dyn_obj.TryGetPropertyValue("additionalType")
                (pattern_matching_result_2, ra_1, singleton_1) = (None, None, None)
                if match_value_3 is not None:
                    if is_array_like(value_3(match_value_3)):
                        pattern_matching_result_2 = 0
                        ra_1 = value_3(match_value_3)

                    elif str(type(value_3(match_value_3))) == "<class \'str\'>":
                        pattern_matching_result_2 = 1
                        singleton_1 = value_3(match_value_3)

                    else: 
                        pattern_matching_result_2 = 2


                else: 
                    pattern_matching_result_2 = 2

                if pattern_matching_result_2 == 0:
                    return ra_1

                elif pattern_matching_result_2 == 1:
                    return [singleton_1]

                elif pattern_matching_result_2 == 2:
                    return None


            roc: LDNode = LDNode(id, st, _arrow1284())
            dyn_obj.DeepCopyPropertiesTo(roc, None, False)
            def action(ph: PropertyHelper) -> None:
                if (True if (True if (ph.Name == "@id") else (ph.Name == "@type")) else (ph.Name == "additionalType")) if ph.IsDynamic else False:
                    ph.RemoveValue(roc)


            iterate(action, roc.GetDynamicPropertyHelpers())
            return roc

        elif pattern_matching_result_1 == 1:
            return None



LDNode_reflection = _expr1285

def LDNode__ctor_479BCDCF(id: str, schema_type: Array[str], additional_type: Array[str] | None=None, context: LDContext | None=None) -> LDNode:
    return LDNode(id, schema_type, additional_type, context)


__all__ = ["ActivePattern__007CNonStringEnumerable_007C__007C", "DynamicObj_DynamicObj__DynamicObj_HasProperty_Z721C83C5", "LDValue_reflection", "LDRef_reflection", "LDGraph_reflection", "LDNode_reflection"]




from __future__ import annotations
from abc import abstractmethod
import builtins
from typing import (Protocol, Any)
from ...fable_library.list import FSharpList
from ...fable_library.util import (IEnumerable_1, IDisposable)

class TextIOBase(Protocol):
    @abstractmethod
    def readline(self, __size: int) -> str:
        ...

    @abstractmethod
    def readlines(self, __hint: int) -> FSharpList[str]:
        ...

    @abstractmethod
    def tell(self) -> int:
        ...

    @abstractmethod
    def write(self, __s: str) -> int:
        ...

    @abstractmethod
    def writelines(self, __lines: IEnumerable_1[str]) -> None:
        ...


class TextIOWrapper(TextIOBase, IDisposable):
    pass

def print(obj: Any | None=None) -> None:
    builtins.print(obj)


__all__ = ["print"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.Helper.collections_ import (ResizeArray_singleton, Option_fromValueWithDefault, ResizeArray_map, ResizeArray_create, ResizeArray_zip)
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...arctrl_core.Helper.regex import ActivePatterns__007CRegex_007C__007C
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.arc_table import ArcTable
from ...arctrl_core.Table.arc_table_aux import (Unchecked_tryGetCellAt, get_empty_cell_for_header)
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.Table.composite_header import CompositeHeader
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.lab_process import LDLabProcess
from ...arctrl_rocrate.LDTypes.lab_protocol import LDLabProtocol
from ...arctrl_rocrate.LDTypes.person import LDPerson
from ...arctrl_rocrate.LDTypes.sample import LDSample
from ...fable_library.int32 import parse
from ...fable_library.list import (FSharpList, choose, length, empty, map as map_2, sort_by)
from ...fable_library.option import (map as map_1, default_arg, value as value_2)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.seq import (indexed, to_list, filter, try_pick, choose as choose_1, map, delay, append, empty as empty_1, singleton)
from ...fable_library.seq2 import List_groupBy
from ...fable_library.types import (to_string, Array)
from ...fable_library.util import (IEnumerable_1, string_hash, compare_primitives)
from .basic import (BaseTypes_decomposeDefinedTerm_Z2F770004, BaseTypes_composeComponent, BaseTypes_composeParameterValue, BaseTypes_composeFactorValue, BaseTypes_composeCharacteristicValue, BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_composeProcessInput, BaseTypes_composeProcessOutput, BaseTypes_decomposeParameterValue_Z2F770004, BaseTypes_decomposeComponent_Z2F770004, BaseTypes_decomposeCharacteristicValue_Z2F770004, BaseTypes_decomposeFactorValue_Z2F770004, BaseTypes_decomposeProcessInput_Z2F770004, BaseTypes_decomposeProcessOutput_Z2F770004)
from .column_index import (ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4, try_get_index)

def _expr3883() -> TypeInfo:
    return class_type("ARCtrl.Conversion.ProcessConversion", None, ProcessConversion)


class ProcessConversion:
    ...

ProcessConversion_reflection = _expr3883

def ProcessConversion_tryGetProtocolType_Z6839B9E8(pv: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> OntologyAnnotation | None:
    match_value: LDNode | None = LDLabProtocol.try_get_intended_use_as_defined_term(pv, graph, context)
    if match_value is None:
        match_value_1: str | None = LDLabProtocol.try_get_intended_use_as_string(pv, context)
        if match_value_1 is None:
            return None

        else: 
            s: str = match_value_1
            return OntologyAnnotation.create(s)


    else: 
        return BaseTypes_decomposeDefinedTerm_Z2F770004(match_value, context)



def ProcessConversion_composeProcessName(process_name_root: str, i: int) -> str:
    return ((("" + process_name_root) + "_") + str(i)) + ""


def ProcessConversion_decomposeProcessName_Z721C83C5(name: str) -> tuple[str, int | None]:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("(?<name>.+)_(?<num>\\d+)", name)
    if active_pattern_result is not None:
        r: Any = active_pattern_result
        return (get_item(groups(r), "name") or "", parse(get_item(groups(r), "num") or "", 511, False, 32))

    else: 
        return (name, None)



def ProcessConversion_tryComponentGetter(general_i: int, value_i: int, value_header: CompositeHeader) -> Callable[[ArcTable, int], LDNode] | None:
    if value_header.tag == 0:
        def Value(table: ArcTable, general_i: Any=general_i, value_i: Any=value_i, value_header: Any=value_header) -> Callable[[int], LDNode]:
            def _arrow3885(i: int, table: Any=table) -> LDNode:
                def _arrow3884(__unit: None=None) -> CompositeCell:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    return get_empty_cell_for_header(value_header, None) if (match_value is None) else match_value

                c: LDNode = BaseTypes_composeComponent(value_header, _arrow3884())
                ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4(c, value_i)
                return c

            return _arrow3885

        return Value

    else: 
        return None



def ProcessConversion_tryParameterGetter(general_i: int, value_i: int, value_header: CompositeHeader) -> Callable[[ArcTable, int], LDNode] | None:
    if value_header.tag == 3:
        def Value(table: ArcTable, general_i: Any=general_i, value_i: Any=value_i, value_header: Any=value_header) -> Callable[[int], LDNode]:
            def _arrow3887(i: int, table: Any=table) -> LDNode:
                def _arrow3886(__unit: None=None) -> CompositeCell:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    return get_empty_cell_for_header(value_header, None) if (match_value is None) else match_value

                p: LDNode = BaseTypes_composeParameterValue(value_header, _arrow3886())
                ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4(p, value_i)
                return p

            return _arrow3887

        return Value

    else: 
        return None



def ProcessConversion_tryFactorGetter(general_i: int, value_i: int, value_header: CompositeHeader) -> Callable[[ArcTable, int], LDNode] | None:
    if value_header.tag == 2:
        def Value(table: ArcTable, general_i: Any=general_i, value_i: Any=value_i, value_header: Any=value_header) -> Callable[[int], LDNode]:
            def _arrow3889(i: int, table: Any=table) -> LDNode:
                def _arrow3888(__unit: None=None) -> CompositeCell:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    return get_empty_cell_for_header(value_header, None) if (match_value is None) else match_value

                f: LDNode = BaseTypes_composeFactorValue(value_header, _arrow3888())
                ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4(f, value_i)
                return f

            return _arrow3889

        return Value

    else: 
        return None



def ProcessConversion_tryCharacteristicGetter(general_i: int, value_i: int, value_header: CompositeHeader) -> Callable[[ArcTable, int], LDNode] | None:
    if value_header.tag == 1:
        def Value(table: ArcTable, general_i: Any=general_i, value_i: Any=value_i, value_header: Any=value_header) -> Callable[[int], LDNode]:
            def _arrow3891(i: int, table: Any=table) -> LDNode:
                def _arrow3890(__unit: None=None) -> CompositeCell:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    return get_empty_cell_for_header(value_header, None) if (match_value is None) else match_value

                c: LDNode = BaseTypes_composeCharacteristicValue(value_header, _arrow3890())
                ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4(c, value_i)
                return c

            return _arrow3891

        return Value

    else: 
        return None



def ProcessConversion_tryGetProtocolTypeGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], LDNode] | None:
    if header.tag == 4:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], LDNode]:
            def _arrow3893(i: int, table: Any=table) -> LDNode:
                def _arrow3892(__unit: None=None) -> OntologyAnnotation:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    if match_value is None:
                        return OntologyAnnotation()

                    else: 
                        cell: CompositeCell = match_value
                        return cell.AsTerm


                return BaseTypes_composeDefinedTerm_ZDED3A0F(_arrow3892())

            return _arrow3893

        return Value

    else: 
        return None



def ProcessConversion_tryGetProtocolREFGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], str] | None:
    if header.tag == 8:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], str]:
            def _arrow3894(i: int, table: Any=table) -> str:
                match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                if match_value is None:
                    return ""

                else: 
                    cell: CompositeCell = match_value
                    return cell.AsFreeText


            return _arrow3894

        return Value

    else: 
        return None



def ProcessConversion_tryGetProtocolDescriptionGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], str] | None:
    if header.tag == 5:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], str]:
            def _arrow3895(i: int, table: Any=table) -> str:
                match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                if match_value is None:
                    return ""

                else: 
                    cell: CompositeCell = match_value
                    return cell.AsFreeText


            return _arrow3895

        return Value

    else: 
        return None



def ProcessConversion_tryGetProtocolURIGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], str] | None:
    if header.tag == 6:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], str]:
            def _arrow3896(i: int, table: Any=table) -> str:
                match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                if match_value is None:
                    return ""

                else: 
                    cell: CompositeCell = match_value
                    return cell.AsFreeText


            return _arrow3896

        return Value

    else: 
        return None



def ProcessConversion_tryGetProtocolVersionGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], str] | None:
    if header.tag == 7:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], str]:
            def _arrow3897(i: int, table: Any=table) -> str:
                match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                if match_value is None:
                    return ""

                else: 
                    cell: CompositeCell = match_value
                    return cell.AsFreeText


            return _arrow3897

        return Value

    else: 
        return None



def ProcessConversion_tryGetInputGetter(general_i: int, header: CompositeHeader, fs: FileSystem | None=None) -> Callable[[ArcTable, int], LDNode] | None:
    if header.tag == 11:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header, fs: Any=fs) -> Callable[[int], LDNode]:
            def _arrow3899(i: int, table: Any=table) -> LDNode:
                def _arrow3898(__unit: None=None) -> CompositeCell:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    return get_empty_cell_for_header(header, None) if (match_value is None) else match_value

                return BaseTypes_composeProcessInput(header, _arrow3898(), fs)

            return _arrow3899

        return Value

    else: 
        return None



def ProcessConversion_tryGetOutputGetter(general_i: int, header: CompositeHeader, fs: FileSystem | None=None) -> Callable[[ArcTable, int], LDNode] | None:
    if header.tag == 12:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header, fs: Any=fs) -> Callable[[int], LDNode]:
            def _arrow3901(i: int, table: Any=table) -> LDNode:
                def _arrow3900(__unit: None=None) -> CompositeCell:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    return get_empty_cell_for_header(header, None) if (match_value is None) else match_value

                return BaseTypes_composeProcessOutput(header, _arrow3900(), fs)

            return _arrow3901

        return Value

    else: 
        return None



def ProcessConversion_tryGetCommentGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], str] | None:
    if header.tag == 14:
        c: str = header.fields[0]
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], str]:
            def _arrow3903(i: int, table: Any=table) -> str:
                def _arrow3902(__unit: None=None) -> Comment:
                    match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                    if match_value is None:
                        return Comment(c)

                    else: 
                        cell: CompositeCell = match_value
                        return Comment(c, cell.AsFreeText)


                return to_string(_arrow3902())

            return _arrow3903

        return Value

    else: 
        return None



def ProcessConversion_tryGetPerformerGetter(general_i: int, header: CompositeHeader) -> Callable[[ArcTable, int], LDNode] | None:
    if header.tag == 9:
        def Value(table: ArcTable, general_i: Any=general_i, header: Any=header) -> Callable[[int], LDNode]:
            def _arrow3904(i: int, table: Any=table) -> LDNode:
                performer: str
                match_value: CompositeCell | None = Unchecked_tryGetCellAt(general_i, i, table.Values)
                if match_value is None:
                    performer = ""

                else: 
                    cell: CompositeCell = match_value
                    performer = cell.AsFreeText

                return LDPerson.create(performer)

            return _arrow3904

        return Value

    else: 
        return None



def ProcessConversion_getProcessGetter(assay_name: str | None, study_name: str | None, process_name_root: str, headers: IEnumerable_1[CompositeHeader], fs: FileSystem | None=None) -> Callable[[ArcTable, int], LDNode]:
    headers_1: IEnumerable_1[tuple[int, CompositeHeader]] = indexed(headers)
    def predicate(arg: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> bool:
        return arg[1].IsCvParamColumn

    value_headers: FSharpList[tuple[int, tuple[int, CompositeHeader]]] = to_list(indexed(filter(predicate, headers_1)))
    def chooser(tupled_arg: tuple[int, tuple[int, CompositeHeader]], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        _arg: tuple[int, CompositeHeader] = tupled_arg[1]
        return ProcessConversion_tryCharacteristicGetter(_arg[0], tupled_arg[0], _arg[1])

    char_getters: FSharpList[Callable[[ArcTable, int], LDNode]] = choose(chooser, value_headers)
    def chooser_1(tupled_arg_1: tuple[int, tuple[int, CompositeHeader]], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        _arg_1: tuple[int, CompositeHeader] = tupled_arg_1[1]
        return ProcessConversion_tryFactorGetter(_arg_1[0], tupled_arg_1[0], _arg_1[1])

    factor_value_getters: FSharpList[Callable[[ArcTable, int], LDNode]] = choose(chooser_1, value_headers)
    def chooser_2(tupled_arg_2: tuple[int, tuple[int, CompositeHeader]], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        _arg_2: tuple[int, CompositeHeader] = tupled_arg_2[1]
        return ProcessConversion_tryParameterGetter(_arg_2[0], tupled_arg_2[0], _arg_2[1])

    parameter_value_getters: FSharpList[Callable[[ArcTable, int], LDNode]] = choose(chooser_2, value_headers)
    def chooser_3(tupled_arg_3: tuple[int, tuple[int, CompositeHeader]], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        _arg_3: tuple[int, CompositeHeader] = tupled_arg_3[1]
        return ProcessConversion_tryComponentGetter(_arg_3[0], tupled_arg_3[0], _arg_3[1])

    component_getters: FSharpList[Callable[[ArcTable, int], LDNode]] = choose(chooser_3, value_headers)
    def chooser_4(tupled_arg_4: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        return ProcessConversion_tryGetProtocolTypeGetter(tupled_arg_4[0], tupled_arg_4[1])

    protocol_type_getter: Callable[[ArcTable, int], LDNode] | None = try_pick(chooser_4, headers_1)
    def chooser_5(tupled_arg_5: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], str] | None:
        return ProcessConversion_tryGetProtocolREFGetter(tupled_arg_5[0], tupled_arg_5[1])

    protocol_refgetter: Callable[[ArcTable, int], str] | None = try_pick(chooser_5, headers_1)
    def chooser_6(tupled_arg_6: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], str] | None:
        return ProcessConversion_tryGetProtocolDescriptionGetter(tupled_arg_6[0], tupled_arg_6[1])

    protocol_description_getter: Callable[[ArcTable, int], str] | None = try_pick(chooser_6, headers_1)
    def chooser_7(tupled_arg_7: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], str] | None:
        return ProcessConversion_tryGetProtocolURIGetter(tupled_arg_7[0], tupled_arg_7[1])

    protocol_urigetter: Callable[[ArcTable, int], str] | None = try_pick(chooser_7, headers_1)
    def chooser_8(tupled_arg_8: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], str] | None:
        return ProcessConversion_tryGetProtocolVersionGetter(tupled_arg_8[0], tupled_arg_8[1])

    protocol_version_getter: Callable[[ArcTable, int], str] | None = try_pick(chooser_8, headers_1)
    def chooser_9(tupled_arg_9: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        return ProcessConversion_tryGetPerformerGetter(tupled_arg_9[0], tupled_arg_9[1])

    performer_getter: Callable[[ArcTable, int], LDNode] | None = try_pick(chooser_9, headers_1)
    def chooser_10(tupled_arg_10: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], str] | None:
        return ProcessConversion_tryGetCommentGetter(tupled_arg_10[0], tupled_arg_10[1])

    comment_getters: FSharpList[Callable[[ArcTable, int], str]] = to_list(choose_1(chooser_10, headers_1))
    input_getter_1: Callable[[ArcTable, int], Array[LDNode]]
    def chooser_11(tupled_arg_11: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        return ProcessConversion_tryGetInputGetter(tupled_arg_11[0], tupled_arg_11[1], fs)

    match_value: Callable[[ArcTable, int], LDNode] | None = try_pick(chooser_11, headers_1)
    if match_value is None:
        def _arrow3906(table_1: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], Array[LDNode]]:
            def _arrow3905(i_1: int) -> Array[LDNode]:
                def mapping_1(f_1: Callable[[ArcTable, int], LDNode]) -> LDNode:
                    return f_1(table_1)(i_1)

                chars_1: Array[LDNode] = list(map(mapping_1, char_getters))
                return ResizeArray_singleton(LDSample.create_sample(((("" + process_name_root) + "_Input_") + str(i_1)) + "", None, chars_1))

            return _arrow3905

        def _arrow3908(table_2: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], Array[LDNode]]:
            def _arrow3907(i_2: int) -> Array[LDNode]:
                return []

            return _arrow3907

        input_getter_1 = _arrow3906 if (length(char_getters) != 0) else _arrow3908

    else: 
        input_getter: Callable[[ArcTable, int], LDNode] = match_value
        def _arrow3910(table: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], Array[LDNode]]:
            def _arrow3909(i: int) -> Array[LDNode]:
                def mapping(f: Callable[[ArcTable, int], LDNode]) -> LDNode:
                    return f(table)(i)

                chars: Array[LDNode] = list(map(mapping, char_getters))
                input: LDNode = input_getter(table)(i)
                if len(chars) > 0:
                    LDSample.set_additional_properties(input, chars)

                return ResizeArray_singleton(input)

            return _arrow3909

        input_getter_1 = _arrow3910

    output_getter_1: Callable[[ArcTable, int], Array[LDNode]]
    def chooser_12(tupled_arg_12: tuple[int, CompositeHeader], assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[ArcTable, int], LDNode] | None:
        return ProcessConversion_tryGetOutputGetter(tupled_arg_12[0], tupled_arg_12[1], fs)

    match_value_1: Callable[[ArcTable, int], LDNode] | None = try_pick(chooser_12, headers_1)
    if match_value_1 is None:
        def _arrow3912(table_4: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], Array[LDNode]]:
            def _arrow3911(i_4: int) -> Array[LDNode]:
                def mapping_3(f_3: Callable[[ArcTable, int], LDNode]) -> LDNode:
                    return f_3(table_4)(i_4)

                factors_1: Array[LDNode] = list(map(mapping_3, factor_value_getters))
                return ResizeArray_singleton(LDSample.create_sample(((("" + process_name_root) + "_Output_") + str(i_4)) + "", None, factors_1))

            return _arrow3911

        def _arrow3914(table_5: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], Array[LDNode]]:
            def _arrow3913(i_5: int) -> Array[LDNode]:
                return []

            return _arrow3913

        output_getter_1 = _arrow3912 if (length(factor_value_getters) != 0) else _arrow3914

    else: 
        output_getter: Callable[[ArcTable, int], LDNode] = match_value_1
        def _arrow3916(table_3: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], Array[LDNode]]:
            def _arrow3915(i_3: int) -> Array[LDNode]:
                def mapping_2(f_2: Callable[[ArcTable, int], LDNode]) -> LDNode:
                    return f_2(table_3)(i_3)

                factors: Array[LDNode] = list(map(mapping_2, factor_value_getters))
                output: LDNode = output_getter(table_3)(i_3)
                if len(factors) > 0:
                    LDSample.set_additional_properties(output, factors)

                return ResizeArray_singleton(output)

            return _arrow3915

        output_getter_1 = _arrow3916

    def _arrow3918(table_6: ArcTable, assay_name: Any=assay_name, study_name: Any=study_name, process_name_root: Any=process_name_root, headers: Any=headers, fs: Any=fs) -> Callable[[int], LDNode]:
        def _arrow3917(i_6: int) -> LDNode:
            pn: str = process_name_root if (table_6.RowCount == 1) else ProcessConversion_composeProcessName(process_name_root, i_6)
            def mapping_4(f_4: Callable[[ArcTable, int], LDNode]) -> LDNode:
                return f_4(table_6)(i_6)

            paramvalues: Array[LDNode] | None = map_1(list, Option_fromValueWithDefault(empty(), map_2(mapping_4, parameter_value_getters)))
            def mapping_6(f_5: Callable[[ArcTable, int], str]) -> str:
                return f_5(table_6)(i_6)

            comments: Array[str] | None = map_1(list, Option_fromValueWithDefault(empty(), map_2(mapping_6, comment_getters)))
            def mapping_8(f_6: Callable[[ArcTable, int], LDNode]) -> LDNode:
                return f_6(table_6)(i_6)

            components: Array[LDNode] | None = map_1(list, Option_fromValueWithDefault(empty(), map_2(mapping_8, component_getters)))
            id: str = LDLabProcess.gen_id(process_name_root, assay_name, study_name) + (("_" + str(i_6)) + "")
            protocol: LDNode | None
            def mapping_10(f_7: Callable[[ArcTable, int], str]) -> str:
                return f_7(table_6)(i_6)

            name: str | None = map_1(mapping_10, protocol_refgetter)
            protocol_id: str = LDLabProtocol.gen_id(name, process_name_root)
            def mapping_11(f_8: Callable[[ArcTable, int], str]) -> str:
                return f_8(table_6)(i_6)

            def mapping_12(f_9: Callable[[ArcTable, int], LDNode]) -> LDNode:
                return f_9(table_6)(i_6)

            def mapping_13(f_10: Callable[[ArcTable, int], str]) -> str:
                return f_10(table_6)(i_6)

            def mapping_14(f_11: Callable[[ArcTable, int], str]) -> str:
                return f_11(table_6)(i_6)

            protocol = LDLabProtocol.create(protocol_id, name, map_1(mapping_11, protocol_description_getter), map_1(mapping_12, protocol_type_getter), None, None, components, None, map_1(mapping_13, protocol_urigetter), map_1(mapping_14, protocol_version_getter))
            match_value: Array[LDNode] = input_getter_1(table_6)(i_6)
            match_value_1: Array[LDNode] = output_getter_1(table_6)(i_6)
            def mapping_15(f_12: Callable[[ArcTable, int], LDNode]) -> LDNode:
                return f_12(table_6)(i_6)

            agent: LDNode | None = map_1(mapping_15, performer_getter)
            return LDLabProcess.create(pn, match_value, match_value_1, id, agent, protocol, paramvalues, None, comments)

        return _arrow3917

    return _arrow3918


def ProcessConversion_groupProcesses_Z27F0B586(processes: FSharpList[LDNode], graph: LDGraph | None=None, context: LDContext | None=None) -> FSharpList[tuple[str, FSharpList[LDNode]]]:
    def projection(p: LDNode, processes: Any=processes, graph: Any=graph, context: Any=context) -> str:
        match_value: str | None = LDLabProcess.try_get_name_as_string(p, context)
        match_value_1: LDNode | None = LDLabProcess.try_get_executes_lab_protocol(p, graph, context)
        (pattern_matching_result, name_1, protocol_2, name_2, protocol_3) = (None, None, None, None, None)
        if match_value is not None:
            if ProcessConversion_decomposeProcessName_Z721C83C5(match_value)[1] is not None:
                pattern_matching_result = 0
                name_1 = match_value

            elif match_value_1 is not None:
                def _arrow3920(__unit: None=None, p: Any=p) -> bool:
                    protocol: LDNode = match_value_1
                    return LDLabProtocol.try_get_name_as_string(protocol, context) is not None

                if _arrow3920():
                    pattern_matching_result = 1
                    protocol_2 = match_value_1

                else: 
                    pattern_matching_result = 2
                    name_2 = match_value


            else: 
                pattern_matching_result = 2
                name_2 = match_value


        elif match_value_1 is not None:
            def _arrow3921(__unit: None=None, p: Any=p) -> bool:
                protocol_1: LDNode = match_value_1
                return LDLabProtocol.try_get_name_as_string(protocol_1, context) is not None

            if _arrow3921():
                pattern_matching_result = 1
                protocol_2 = match_value_1

            else: 
                pattern_matching_result = 3
                protocol_3 = match_value_1


        else: 
            pattern_matching_result = 4

        if pattern_matching_result == 0:
            return ProcessConversion_decomposeProcessName_Z721C83C5(name_1)[0]

        elif pattern_matching_result == 1:
            return default_arg(LDLabProtocol.try_get_name_as_string(protocol_2, context), "")

        elif pattern_matching_result == 2:
            return name_2

        elif pattern_matching_result == 3:
            return protocol_3.Id

        elif pattern_matching_result == 4:
            return create_missing_identifier()


    class ObjectExpr3923:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3922(x: str, y: str) -> bool:
                return x == y

            return _arrow3922

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return List_groupBy(projection, processes, ObjectExpr3923())


def ProcessConversion_processToRows_Z6839B9E8(p: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[FSharpList[tuple[CompositeHeader, CompositeCell]]]:
    def f(ppv: LDNode, p: Any=p, graph: Any=graph, context: Any=context) -> tuple[tuple[CompositeHeader, CompositeCell], int | None]:
        return (BaseTypes_decomposeParameterValue_Z2F770004(ppv, context), try_get_index(ppv))

    pvs: Array[tuple[tuple[CompositeHeader, CompositeCell], int | None]] = ResizeArray_map(f, LDLabProcess.get_parameter_values(p, graph, context))
    components: Array[tuple[tuple[CompositeHeader, CompositeCell], int | None]]
    match_value: LDNode | None = LDLabProcess.try_get_executes_lab_protocol(p, graph, context)
    if match_value is None:
        components = []

    else: 
        prot: LDNode = match_value
        def f_1(ppv_1: LDNode, p: Any=p, graph: Any=graph, context: Any=context) -> tuple[tuple[CompositeHeader, CompositeCell], int | None]:
            return (BaseTypes_decomposeComponent_Z2F770004(ppv_1, context), try_get_index(ppv_1))

        components = ResizeArray_map(f_1, LDLabProtocol.get_components(prot, graph, context))

    prot_vals: FSharpList[tuple[CompositeHeader, CompositeCell]]
    match_value_1: LDNode | None = LDLabProcess.try_get_executes_lab_protocol(p, graph, context)
    if match_value_1 is None:
        prot_vals = empty()

    else: 
        prot_1: LDNode = match_value_1
        def _arrow3933(__unit: None=None, p: Any=p, graph: Any=graph, context: Any=context) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
            def _arrow3925(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                match_value_2: str | None = LDLabProtocol.try_get_name_as_string(prot_1, context)
                if match_value_2 is None:
                    return empty_1()

                else: 
                    return singleton((CompositeHeader(8), CompositeCell(1, match_value_2)))


            def _arrow3932(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                def _arrow3926(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                    match_value_3: str | None = LDLabProtocol.try_get_description_as_string(prot_1, context)
                    if match_value_3 is None:
                        return empty_1()

                    else: 
                        return singleton((CompositeHeader(5), CompositeCell(1, match_value_3)))


                def _arrow3931(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                    def _arrow3927(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                        match_value_4: str | None = LDLabProtocol.try_get_url(prot_1, context)
                        if match_value_4 is None:
                            return empty_1()

                        else: 
                            return singleton((CompositeHeader(6), CompositeCell(1, match_value_4)))


                    def _arrow3930(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                        def _arrow3928(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                            match_value_5: str | None = LDLabProtocol.try_get_version_as_string(prot_1, context)
                            if match_value_5 is None:
                                return empty_1()

                            else: 
                                return singleton((CompositeHeader(7), CompositeCell(1, match_value_5)))


                        def _arrow3929(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                            match_value_6: OntologyAnnotation | None = ProcessConversion_tryGetProtocolType_Z6839B9E8(prot_1, graph, context)
                            if match_value_6 is None:
                                return empty_1()

                            else: 
                                return singleton((CompositeHeader(4), CompositeCell(0, match_value_6)))


                        return append(_arrow3928(), delay(_arrow3929))

                    return append(_arrow3927(), delay(_arrow3930))

                return append(_arrow3926(), delay(_arrow3931))

            return append(_arrow3925(), delay(_arrow3932))

        prot_vals = to_list(delay(_arrow3933))

    def f_2(c: str, p: Any=p, graph: Any=graph, context: Any=context) -> tuple[CompositeHeader, CompositeCell]:
        c_1: Comment = Comment.from_string(c)
        return (CompositeHeader(14, default_arg(c_1.Name, "")), CompositeCell(1, default_arg(c_1.Value, "")))

    comments: Array[tuple[CompositeHeader, CompositeCell]] = ResizeArray_map(f_2, LDLabProcess.get_disambiguating_descriptions_as_string(p, context))
    inputs: Array[LDNode] = LDLabProcess.get_objects(p, graph, context)
    outputs: Array[LDNode] = LDLabProcess.get_results(p, graph, context)
    def _arrow3934(Value: LDNode, p: Any=p, graph: Any=graph, context: Any=context) -> LDNode | None:
        return Value

    def _arrow3935(Value_1: LDNode, p: Any=p, graph: Any=graph, context: Any=context) -> LDNode | None:
        return Value_1

    def _arrow3936(Value_2: LDNode, p: Any=p, graph: Any=graph, context: Any=context) -> LDNode | None:
        return Value_2

    def _arrow3937(Value_3: LDNode, p: Any=p, graph: Any=graph, context: Any=context) -> LDNode | None:
        return Value_3

    pattern_input: tuple[Array[LDNode | None], Array[LDNode | None]] = ((ResizeArray_create(len(outputs), None), ResizeArray_map(_arrow3934, outputs))) if ((len(outputs) != 0) if (len(inputs) == 0) else False) else (((ResizeArray_map(_arrow3935, inputs), ResizeArray_create(len(inputs), None))) if ((len(outputs) == 0) if (len(inputs) != 0) else False) else ((ResizeArray_map(_arrow3936, inputs), ResizeArray_map(_arrow3937, outputs))))
    outputs_1: Array[LDNode | None] = pattern_input[1]
    inputs_1: Array[LDNode | None] = pattern_input[0]
    if (len(outputs_1) == 0) if (len(inputs_1) == 0) else False:
        def mapping(tuple_1: tuple[tuple[CompositeHeader, CompositeCell], int | None], p: Any=p, graph: Any=graph, context: Any=context) -> tuple[CompositeHeader, CompositeCell]:
            return tuple_1[0]

        def projection(arg: tuple[tuple[CompositeHeader, CompositeCell], int | None], p: Any=p, graph: Any=graph, context: Any=context) -> int:
            return default_arg(arg[1], 10000)

        def _arrow3939(__unit: None=None, p: Any=p, graph: Any=graph, context: Any=context) -> IEnumerable_1[tuple[tuple[CompositeHeader, CompositeCell], int | None]]:
            def _arrow3938(__unit: None=None) -> IEnumerable_1[tuple[tuple[CompositeHeader, CompositeCell], int | None]]:
                return pvs

            return append(components, delay(_arrow3938))

        class ObjectExpr3940:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        vals: FSharpList[tuple[CompositeHeader, CompositeCell]] = map_2(mapping, sort_by(projection, to_list(delay(_arrow3939)), ObjectExpr3940()))
        def _arrow3943(__unit: None=None, p: Any=p, graph: Any=graph, context: Any=context) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
            def _arrow3942(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                def _arrow3941(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                    return comments

                return append(vals, delay(_arrow3941))

            return append(prot_vals, delay(_arrow3942))

        return ResizeArray_singleton(to_list(delay(_arrow3943)))

    else: 
        def f_5(tupled_arg: tuple[LDNode | None, LDNode | None], p: Any=p, graph: Any=graph, context: Any=context) -> FSharpList[tuple[CompositeHeader, CompositeCell]]:
            i: LDNode | None = tupled_arg[0]
            o: LDNode | None = tupled_arg[1]
            chars: Array[tuple[tuple[CompositeHeader, CompositeCell], int | None]]
            if i is None:
                chars = []

            else: 
                i_1: LDNode = i
                def f_3(cv: LDNode, tupled_arg: Any=tupled_arg) -> tuple[tuple[CompositeHeader, CompositeCell], int | None]:
                    return (BaseTypes_decomposeCharacteristicValue_Z2F770004(cv, context), try_get_index(cv))

                chars = ResizeArray_map(f_3, LDSample.get_characteristics(i_1, graph, context))

            factors: Array[tuple[tuple[CompositeHeader, CompositeCell], int | None]]
            if o is None:
                factors = []

            else: 
                o_1: LDNode = o
                def f_4(fv: LDNode, tupled_arg: Any=tupled_arg) -> tuple[tuple[CompositeHeader, CompositeCell], int | None]:
                    return (BaseTypes_decomposeFactorValue_Z2F770004(fv, context), try_get_index(fv))

                factors = ResizeArray_map(f_4, LDSample.get_factors(o_1, graph, context))

            def mapping_1(tuple_3: tuple[tuple[CompositeHeader, CompositeCell], int | None], tupled_arg: Any=tupled_arg) -> tuple[CompositeHeader, CompositeCell]:
                return tuple_3[0]

            def projection_1(arg_1: tuple[tuple[CompositeHeader, CompositeCell], int | None], tupled_arg: Any=tupled_arg) -> int:
                return default_arg(arg_1[1], 10000)

            def _arrow3947(__unit: None=None, tupled_arg: Any=tupled_arg) -> IEnumerable_1[tuple[tuple[CompositeHeader, CompositeCell], int | None]]:
                def _arrow3946(__unit: None=None) -> IEnumerable_1[tuple[tuple[CompositeHeader, CompositeCell], int | None]]:
                    def _arrow3945(__unit: None=None) -> IEnumerable_1[tuple[tuple[CompositeHeader, CompositeCell], int | None]]:
                        def _arrow3944(__unit: None=None) -> IEnumerable_1[tuple[tuple[CompositeHeader, CompositeCell], int | None]]:
                            return factors

                        return append(pvs, delay(_arrow3944))

                    return append(components, delay(_arrow3945))

                return append(chars, delay(_arrow3946))

            class ObjectExpr3948:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            vals_1: FSharpList[tuple[CompositeHeader, CompositeCell]] = map_2(mapping_1, sort_by(projection_1, to_list(delay(_arrow3947)), ObjectExpr3948()))
            def _arrow3953(__unit: None=None, tupled_arg: Any=tupled_arg) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                def _arrow3952(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                    def _arrow3951(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                        def _arrow3950(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                            def _arrow3949(__unit: None=None) -> IEnumerable_1[tuple[CompositeHeader, CompositeCell]]:
                                return singleton(BaseTypes_decomposeProcessOutput_Z2F770004(value_2(o), context)) if (o is not None) else empty_1()

                            return append(comments, delay(_arrow3949))

                        return append(vals_1, delay(_arrow3950))

                    return append(prot_vals, delay(_arrow3951))

                return append(singleton(BaseTypes_decomposeProcessInput_Z2F770004(value_2(i), context)) if (i is not None) else empty_1(), delay(_arrow3952))

            return to_list(delay(_arrow3953))

        return ResizeArray_map(f_5, ResizeArray_zip(inputs_1, outputs_1))



__all__ = ["ProcessConversion_reflection", "ProcessConversion_tryGetProtocolType_Z6839B9E8", "ProcessConversion_composeProcessName", "ProcessConversion_decomposeProcessName_Z721C83C5", "ProcessConversion_tryComponentGetter", "ProcessConversion_tryParameterGetter", "ProcessConversion_tryFactorGetter", "ProcessConversion_tryCharacteristicGetter", "ProcessConversion_tryGetProtocolTypeGetter", "ProcessConversion_tryGetProtocolREFGetter", "ProcessConversion_tryGetProtocolDescriptionGetter", "ProcessConversion_tryGetProtocolURIGetter", "ProcessConversion_tryGetProtocolVersionGetter", "ProcessConversion_tryGetInputGetter", "ProcessConversion_tryGetOutputGetter", "ProcessConversion_tryGetCommentGetter", "ProcessConversion_tryGetPerformerGetter", "ProcessConversion_getProcessGetter", "ProcessConversion_groupProcesses_Z27F0B586", "ProcessConversion_processToRows_Z6839B9E8"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1734() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Person.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Person", string_type), ("first_name", string_type), ("last_name", string_type), ("mid_initials", string_type), ("email", string_type), ("address", string_type), ("phone", string_type), ("fax", string_type), ("comments", string_type), ("roles", string_type), ("affiliation", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Person: str
    first_name: str
    last_name: str
    mid_initials: str
    email: str
    address: str
    phone: str
    fax: str
    comments: str
    roles: str
    affiliation: str

IContext_reflection = _expr1734

def _arrow1749(__unit: None=None) -> IEncodable:
    class ObjectExpr1735(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1736(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:Person")

    class ObjectExpr1737(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:identifier")

    class ObjectExpr1738(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:givenName")

    class ObjectExpr1739(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:familyName")

    class ObjectExpr1740(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:additionalName")

    class ObjectExpr1741(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:email")

    class ObjectExpr1742(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:address")

    class ObjectExpr1743(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:telephone")

    class ObjectExpr1744(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            return helpers_9.encode_string("sdo:faxNumber")

    class ObjectExpr1745(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
            return helpers_10.encode_string("sdo:disambiguatingDescription")

    class ObjectExpr1746(IEncodable):
        def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
            return helpers_11.encode_string("sdo:jobTitle")

    class ObjectExpr1747(IEncodable):
        def Encode(self, helpers_12: IEncoderHelpers_1[Any]) -> Any:
            return helpers_12.encode_string("sdo:affiliation")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1735()), ("Person", ObjectExpr1736()), ("orcid", ObjectExpr1737()), ("firstName", ObjectExpr1738()), ("lastName", ObjectExpr1739()), ("midInitials", ObjectExpr1740()), ("email", ObjectExpr1741()), ("address", ObjectExpr1742()), ("phone", ObjectExpr1743()), ("fax", ObjectExpr1744()), ("comments", ObjectExpr1745()), ("roles", ObjectExpr1746()), ("affiliation", ObjectExpr1747())])
    class ObjectExpr1748(IEncodable):
        def Encode(self, helpers_13: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_13))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_13.encode_object(arg)

    return ObjectExpr1748()


context_jsonvalue: IEncodable = _arrow1749()

def _arrow1754(__unit: None=None) -> IEncodable:
    class ObjectExpr1750(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1751(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:Person")

    class ObjectExpr1752(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:name")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1750()), ("Person", ObjectExpr1751()), ("name", ObjectExpr1752())])
    class ObjectExpr1753(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr1753()


context_minimal_json_value: IEncodable = _arrow1754()

context_str: str = "\r\n{\r\n  \"@context\": {\r\n    \"sdo\": \"http://schema.org/\",\r\n    \"arc\": \"http://purl.org/nfdi4plants/ontology/\",\r\n\r\n    \"Person\": \"sdo:Person\",\r\n    \"firstName\": \"sdo:givenName\",\r\n    \"lastName\": \"sdo:familyName\",\r\n    \"midInitials\": \"sdo:additionalName\",\r\n    \"email\": \"sdo:email\",\r\n    \"address\": \"sdo:address\",\r\n    \"phone\": \"sdo:telephone\",\r\n    \"fax\": \"sdo:faxNumber\",\r\n    \"comments\": \"sdo:disambiguatingDescription\",\r\n    \"roles\": \"sdo:jobTitle\",\r\n    \"affiliation\": \"sdo:affiliation\"\r\n  }\r\n}\r\n    "

__all__ = ["IContext_reflection", "context_jsonvalue", "context_minimal_json_value", "context_str"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.int32 import parse
from ..fable_library.list import (cons, FSharpList, is_empty, empty, head, singleton as singleton_1, tail, of_array_with_tail, reverse, append as append_1)
from ..fable_library.map_util import get_item_from_dict
from ..fable_library.option import (map, value as value_2)
from ..fable_library.reg_exp import (match, create, get_item, groups)
from ..fable_library.seq import (to_list, delay, append, singleton, take_while, skip_while, map as map_1, collect, empty as empty_1)
from ..fable_library.string_ import (to_fail, printf, split as split_2, trim_end)
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from .preprocessing import read as read_1
from .regex_active_patterns import (_007CSequenceMinusOpener_007C__007C, _007CYamlComment_007C__007C, _007CKeyValue_007C__007C, _007CYamlValue_007C__007C, _007CKey_007C__007C, _007CInlineJSON_007C__007C, _007CJSONKeyOpener_007C__007C, _007CJSONCloser_007C__007C, _007CInlineSequence_007C__007C, _007CSequenceSquareOpener_007C__007C, _007CSequenceSquareCloser_007C__007C, _007CSchemaNamespace_007C__007C)
from .yamlicious_types import (PreprocessorElement, YAMLElement, YAMLContent_create_27AED5E3, Preprocessor)

def restore_string_replace(string_dict: Any, v: str) -> str:
    m: Any = match(create("\\<s f=(?P<index>\\d+)\\/\\>"), v)
    if m is not None:
        return get_item_from_dict(string_dict, parse(get_item(groups(m), "index") or "", 511, False, 32))

    else: 
        return v



def restore_comment_replace(comment_dict: Any, comment_id: int | None=None) -> str | None:
    def mapping(id: int, comment_dict: Any=comment_dict, comment_id: Any=comment_id) -> str:
        return get_item_from_dict(comment_dict, id)

    return map(mapping, comment_id)


def collect_sequence_elements(eles: FSharpList[PreprocessorElement]) -> FSharpList[FSharpList[PreprocessorElement]]:
    (pattern_matching_result, rest, v, yaml_ast_list) = (None, None, None, None)
    if not is_empty(eles):
        active_pattern_result: dict[str, Any] | None = _007CSequenceMinusOpener_007C__007C(head(eles))
        if active_pattern_result is not None:
            if not is_empty(tail(eles)):
                if head(tail(eles)).tag == 1:
                    pattern_matching_result = 0
                    rest = tail(tail(eles))
                    v = active_pattern_result
                    yaml_ast_list = head(tail(eles)).fields[0]

                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def _arrow413(__unit: None=None, eles: Any=eles) -> IEnumerable_1[FSharpList[PreprocessorElement]]:
            def _arrow412(__unit: None=None) -> IEnumerable_1[FSharpList[PreprocessorElement]]:
                return collect_sequence_elements(rest)

            return append(singleton(cons(PreprocessorElement(2, value_2(v["Value"])), yaml_ast_list)) if (v["Value"] is not None) else singleton(yaml_ast_list), delay(_arrow412))

        return to_list(delay(_arrow413))

    elif pattern_matching_result == 1:
        if is_empty(eles):
            return empty()

        else: 
            active_pattern_result_1: dict[str, Any] | None = _007CSequenceMinusOpener_007C__007C(head(eles))
            if active_pattern_result_1 is not None:
                v_1: dict[str, Any] = active_pattern_result_1
                def _arrow415(__unit: None=None, eles: Any=eles) -> IEnumerable_1[FSharpList[PreprocessorElement]]:
                    def _arrow414(__unit: None=None) -> IEnumerable_1[FSharpList[PreprocessorElement]]:
                        return collect_sequence_elements(tail(eles))

                    return append(singleton(singleton_1(PreprocessorElement(2, value_2(v_1["Value"])))), delay(_arrow414))

                return to_list(delay(_arrow415))

            elif _007CYamlComment_007C__007C(head(eles)) is not None:
                def _arrow417(__unit: None=None, eles: Any=eles) -> IEnumerable_1[FSharpList[PreprocessorElement]]:
                    def _arrow416(__unit: None=None) -> IEnumerable_1[FSharpList[PreprocessorElement]]:
                        return collect_sequence_elements(tail(eles))

                    return append(singleton(singleton_1(head(eles))), delay(_arrow416))

                return to_list(delay(_arrow417))

            else: 
                return to_fail(printf("Unknown pattern for sequence elements: %A"))(eles)





def is_sequence_element(e: PreprocessorElement) -> bool:
    (pattern_matching_result,) = (None,)
    if e.tag == 1:
        pattern_matching_result = 0

    elif _007CSequenceMinusOpener_007C__007C(e) is not None:
        pattern_matching_result = 0

    elif _007CYamlComment_007C__007C(e) is not None:
        pattern_matching_result = 0

    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return True

    elif pattern_matching_result == 1:
        return False



def tokenize(yaml_list: FSharpList[PreprocessorElement], string_dict: Any, comment_dict: Any) -> YAMLElement:
    def loop_read(restlist_mut: FSharpList[PreprocessorElement], acc_mut: FSharpList[YAMLElement], yaml_list: Any=yaml_list, string_dict: Any=string_dict, comment_dict: Any=comment_dict) -> YAMLElement:
        while True:
            (restlist, acc) = (restlist_mut, acc_mut)
            (pattern_matching_result, rest0, v, yaml_ast_list) = (None, None, None, None)
            if not is_empty(restlist):
                active_pattern_result: dict[str, Any] | None = _007CSchemaNamespace_007C__007C(head(restlist))
                if active_pattern_result is not None:
                    if not is_empty(tail(restlist)):
                        if head(tail(restlist)).tag == 1:
                            pattern_matching_result = 0
                            rest0 = tail(tail(restlist))
                            v = active_pattern_result
                            yaml_ast_list = head(tail(restlist)).fields[0]

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                def _arrow418(e: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                    return is_sequence_element(e)

                sequence_elements: FSharpList[FSharpList[PreprocessorElement]] = collect_sequence_elements(to_list(take_while(_arrow418, rest0)))
                def _arrow419(e_1: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                    return is_sequence_element(e_1)

                restlist_mut = to_list(skip_while(_arrow419, rest0))
                def _arrow425(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                    def _arrow424(__unit: None=None) -> IEnumerable_1[YAMLElement]:
                        def _arrow423(i: FSharpList[PreprocessorElement]) -> YAMLElement:
                            return loop_read(i, empty())

                        return map_1(_arrow423, sequence_elements)

                    return append(singleton(loop_read(cons(PreprocessorElement(2, v["Key"]), yaml_ast_list), empty())), delay(_arrow424))

                acc_mut = cons(YAMLElement(2, to_list(delay(_arrow425))), acc)
                continue

            elif pattern_matching_result == 1:
                (pattern_matching_result_1, rest0_1, v_1, rest0_2, v_2, yaml_ast_list_1) = (None, None, None, None, None, None)
                if not is_empty(restlist):
                    active_pattern_result_1: dict[str, Any] | None = _007CSchemaNamespace_007C__007C(head(restlist))
                    if active_pattern_result_1 is not None:
                        pattern_matching_result_1 = 0
                        rest0_1 = tail(restlist)
                        v_1 = active_pattern_result_1

                    else: 
                        active_pattern_result_2: dict[str, Any] | None = _007CSequenceMinusOpener_007C__007C(head(restlist))
                        if active_pattern_result_2 is not None:
                            if not is_empty(tail(restlist)):
                                if head(tail(restlist)).tag == 1:
                                    pattern_matching_result_1 = 1
                                    rest0_2 = tail(tail(restlist))
                                    v_2 = active_pattern_result_2
                                    yaml_ast_list_1 = head(tail(restlist)).fields[0]

                                else: 
                                    pattern_matching_result_1 = 2


                            else: 
                                pattern_matching_result_1 = 2


                        else: 
                            pattern_matching_result_1 = 2



                else: 
                    pattern_matching_result_1 = 2

                if pattern_matching_result_1 == 0:
                    def _arrow426(e_2: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                        return is_sequence_element(e_2)

                    sequence_elements_1: FSharpList[FSharpList[PreprocessorElement]] = collect_sequence_elements(to_list(take_while(_arrow426, rest0_1)))
                    def _arrow427(e_3: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                        return is_sequence_element(e_3)

                    restlist_mut = to_list(skip_while(_arrow427, rest0_1))
                    def _arrow433(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                        def _arrow432(__unit: None=None) -> IEnumerable_1[YAMLElement]:
                            def _arrow431(i_1: FSharpList[PreprocessorElement]) -> YAMLElement:
                                return loop_read(i_1, empty())

                            return map_1(_arrow431, sequence_elements_1)

                        return append(singleton(loop_read(singleton_1(PreprocessorElement(2, v_1["Key"])), empty())), delay(_arrow432))

                    acc_mut = cons(YAMLElement(2, to_list(delay(_arrow433))), acc)
                    continue

                elif pattern_matching_result_1 == 1:
                    object_list_1: FSharpList[PreprocessorElement] = cons(PreprocessorElement(2, value_2(v_2["Value"])), yaml_ast_list_1) if (v_2["Value"] is not None) else yaml_ast_list_1
                    def _arrow434(e_4: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                        return is_sequence_element(e_4)

                    sequence_elements_2: FSharpList[FSharpList[PreprocessorElement]] = collect_sequence_elements(to_list(take_while(_arrow434, rest0_2)))
                    def _arrow435(e_5: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                        return is_sequence_element(e_5)

                    restlist_mut = to_list(skip_while(_arrow435, rest0_2))
                    def _arrow441(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                        def _arrow440(__unit: None=None) -> IEnumerable_1[YAMLElement]:
                            def _arrow439(i_2: FSharpList[PreprocessorElement]) -> YAMLElement:
                                return loop_read(i_2, empty())

                            return map_1(_arrow439, sequence_elements_2)

                        return append(singleton(loop_read(object_list_1, empty())), delay(_arrow440))

                    acc_mut = cons(YAMLElement(2, to_list(delay(_arrow441))), acc)
                    continue

                elif pattern_matching_result_1 == 2:
                    (pattern_matching_result_2, rest0_3, v_3, rest_4, v_4, closer, i_list, opener, rest_5) = (None, None, None, None, None, None, None, None, None)
                    if not is_empty(restlist):
                        active_pattern_result_3: dict[str, Any] | None = _007CSequenceMinusOpener_007C__007C(head(restlist))
                        if active_pattern_result_3 is not None:
                            pattern_matching_result_2 = 0
                            rest0_3 = tail(restlist)
                            v_3 = active_pattern_result_3

                        else: 
                            active_pattern_result_4: dict[str, Any] | None = _007CInlineSequence_007C__007C(head(restlist))
                            if active_pattern_result_4 is not None:
                                pattern_matching_result_2 = 1
                                rest_4 = tail(restlist)
                                v_4 = active_pattern_result_4

                            else: 
                                active_pattern_result_5: dict[str, Any] | None = _007CSequenceSquareOpener_007C__007C(head(restlist))
                                if active_pattern_result_5 is not None:
                                    if not is_empty(tail(restlist)):
                                        if head(tail(restlist)).tag == 1:
                                            if not is_empty(tail(tail(restlist))):
                                                active_pattern_result_6: dict[str, Any] | None = _007CSequenceSquareCloser_007C__007C(head(tail(tail(restlist))))
                                                if active_pattern_result_6 is not None:
                                                    pattern_matching_result_2 = 2
                                                    closer = active_pattern_result_6
                                                    i_list = head(tail(restlist)).fields[0]
                                                    opener = active_pattern_result_5
                                                    rest_5 = tail(tail(tail(restlist)))

                                                else: 
                                                    pattern_matching_result_2 = 3


                                            else: 
                                                pattern_matching_result_2 = 3


                                        else: 
                                            pattern_matching_result_2 = 3


                                    else: 
                                        pattern_matching_result_2 = 3


                                else: 
                                    pattern_matching_result_2 = 3




                    else: 
                        pattern_matching_result_2 = 3

                    if pattern_matching_result_2 == 0:
                        def _arrow442(e_6: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                            return is_sequence_element(e_6)

                        sequence_elements_3: FSharpList[FSharpList[PreprocessorElement]] = collect_sequence_elements(to_list(take_while(_arrow442, rest0_3)))
                        def _arrow443(e_7: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                            return is_sequence_element(e_7)

                        restlist_mut = to_list(skip_while(_arrow443, rest0_3))
                        def _arrow449(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                            def _arrow448(__unit: None=None) -> IEnumerable_1[YAMLElement]:
                                def _arrow447(i_3: FSharpList[PreprocessorElement]) -> YAMLElement:
                                    return loop_read(i_3, empty())

                                return map_1(_arrow447, sequence_elements_3)

                            return append(singleton(loop_read(singleton_1(PreprocessorElement(2, value_2(v_3["Value"]))), empty())), delay(_arrow448))

                        acc_mut = cons(YAMLElement(2, to_list(delay(_arrow449))), acc)
                        continue

                    elif pattern_matching_result_2 == 1:
                        c: str | None = restore_comment_replace(comment_dict, v_4["Comment"])
                        split: Array[str] = split_2(v_4["Value"], [","], None, 1)
                        def _arrow451(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                            def _arrow450(value: str) -> YAMLElement:
                                return loop_read(singleton_1(PreprocessorElement(2, value.strip())), empty())

                            return map_1(_arrow450, split)

                        current_4: YAMLElement = YAMLElement(2, to_list(delay(_arrow451)))
                        restlist_mut = rest_4
                        acc_mut = of_array_with_tail([current_4, YAMLElement(4, value_2(c))], acc) if (c is not None) else cons(current_4, acc)
                        continue

                    elif pattern_matching_result_2 == 2:
                        c1: str | None = restore_comment_replace(comment_dict, opener["Comment"])
                        c2: str | None = restore_comment_replace(comment_dict, closer["Comment"])
                        def _arrow453(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                            def _arrow452(i_4: PreprocessorElement) -> IEnumerable_1[YAMLElement]:
                                return singleton(loop_read(singleton_1(PreprocessorElement(2, trim_end(i_4.fields[0], ",")) if (i_4.tag == 2) else to_fail(printf("Unexpected element in MultiLineSquareBrackets: %A"))(i_4)), empty()))

                            return collect(_arrow452, i_list)

                        current_5: YAMLElement = YAMLElement(2, to_list(delay(_arrow453)))
                        restlist_mut = rest_5
                        def _arrow454(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> FSharpList[YAMLElement]:
                            c2_2: str = c2
                            return of_array_with_tail([YAMLElement(4, c2_2), current_5], acc)

                        def _arrow455(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> FSharpList[YAMLElement]:
                            c1_2: str = c1
                            return of_array_with_tail([current_5, YAMLElement(4, c1_2)], acc)

                        def _arrow456(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> FSharpList[YAMLElement]:
                            c1_1: str = c1
                            c2_1: str = c2
                            return of_array_with_tail([YAMLElement(4, c2_1), current_5, YAMLElement(4, c1_1)], acc)

                        acc_mut = (cons(current_5, acc) if (c2 is None) else _arrow454()) if (c1 is None) else (_arrow455() if (c2 is None) else _arrow456())
                        continue

                    elif pattern_matching_result_2 == 3:
                        (pattern_matching_result_3, rest_6, v_5, closer_1, i_list_1, opener_1, rest_7) = (None, None, None, None, None, None, None)
                        if not is_empty(restlist):
                            active_pattern_result_7: dict[str, Any] | None = _007CInlineJSON_007C__007C(head(restlist))
                            if active_pattern_result_7 is not None:
                                pattern_matching_result_3 = 0
                                rest_6 = tail(restlist)
                                v_5 = active_pattern_result_7

                            else: 
                                active_pattern_result_8: dict[str, Any] | None = _007CJSONKeyOpener_007C__007C(head(restlist))
                                if active_pattern_result_8 is not None:
                                    if not is_empty(tail(restlist)):
                                        if head(tail(restlist)).tag == 1:
                                            if not is_empty(tail(tail(restlist))):
                                                active_pattern_result_9: dict[str, Any] | None = _007CJSONCloser_007C__007C(head(tail(tail(restlist))))
                                                if active_pattern_result_9 is not None:
                                                    pattern_matching_result_3 = 1
                                                    closer_1 = active_pattern_result_9
                                                    i_list_1 = head(tail(restlist)).fields[0]
                                                    opener_1 = active_pattern_result_8
                                                    rest_7 = tail(tail(tail(restlist)))

                                                else: 
                                                    pattern_matching_result_3 = 2


                                            else: 
                                                pattern_matching_result_3 = 2


                                        else: 
                                            pattern_matching_result_3 = 2


                                    else: 
                                        pattern_matching_result_3 = 2


                                else: 
                                    pattern_matching_result_3 = 2



                        else: 
                            pattern_matching_result_3 = 2

                        if pattern_matching_result_3 == 0:
                            c_1: str | None = restore_comment_replace(comment_dict, v_5["Comment"])
                            split_1: Array[str] = split_2(v_5["Value"], [","], None, 1)
                            def _arrow458(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                                def _arrow457(value_1: str) -> IEnumerable_1[YAMLElement]:
                                    match_value_1: YAMLElement = loop_read(singleton_1(PreprocessorElement(2, value_1.strip())), empty())
                                    if match_value_1.tag == 3:
                                        return match_value_1.fields[0]

                                    else: 
                                        raise Exception("Unexpected element in InlineJSON")
                                        return empty_1()


                                return collect(_arrow457, split_1)

                            current_6: FSharpList[YAMLElement] = reverse(to_list(delay(_arrow458)))
                            restlist_mut = rest_6
                            acc_mut = append_1(current_6, cons(YAMLElement(4, value_2(c_1)), acc)) if (c_1 is not None) else append_1(current_6, acc)
                            continue

                        elif pattern_matching_result_3 == 1:
                            c1_3: str | None = restore_comment_replace(comment_dict, opener_1["Comment"])
                            c2_3: str | None = restore_comment_replace(comment_dict, closer_1["Comment"])
                            def _arrow460(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                                def _arrow459(i_5: PreprocessorElement) -> IEnumerable_1[YAMLElement]:
                                    match_value_2: YAMLElement = loop_read(singleton_1(PreprocessorElement(2, trim_end(i_5.fields[0], ",")) if (i_5.tag == 2) else to_fail(printf("Unexpected element in MultiLineSquareBrackets: %A"))(i_5)), empty())
                                    if match_value_2.tag == 3:
                                        return match_value_2.fields[0]

                                    else: 
                                        raise Exception("Unexpected element in MultilineJSON")
                                        return empty_1()


                                return collect(_arrow459, i_list_1)

                            current_7: YAMLElement = YAMLElement(0, YAMLContent_create_27AED5E3(opener_1["Key"], c1_3), YAMLElement(3, to_list(delay(_arrow460))))
                            restlist_mut = rest_7
                            acc_mut = cons(current_7, acc) if (c2_3 is None) else of_array_with_tail([YAMLElement(4, c2_3), current_7], acc)
                            continue

                        elif pattern_matching_result_3 == 2:
                            (pattern_matching_result_4, rest_8, v_6, yaml_ast_list_2) = (None, None, None, None)
                            if not is_empty(restlist):
                                active_pattern_result_10: dict[str, Any] | None = _007CKey_007C__007C(head(restlist))
                                if active_pattern_result_10 is not None:
                                    if not is_empty(tail(restlist)):
                                        if head(tail(restlist)).tag == 1:
                                            pattern_matching_result_4 = 0
                                            rest_8 = tail(tail(restlist))
                                            v_6 = active_pattern_result_10
                                            yaml_ast_list_2 = head(tail(restlist)).fields[0]

                                        else: 
                                            pattern_matching_result_4 = 1


                                    else: 
                                        pattern_matching_result_4 = 1


                                else: 
                                    pattern_matching_result_4 = 1


                            else: 
                                pattern_matching_result_4 = 1

                            if pattern_matching_result_4 == 0:
                                restlist_mut = rest_8
                                acc_mut = cons(YAMLElement(0, YAMLContent_create_27AED5E3(v_6["Key"], restore_comment_replace(comment_dict, v_6["Comment"])), loop_read(yaml_ast_list_2, empty())), acc)
                                continue

                            elif pattern_matching_result_4 == 1:
                                (pattern_matching_result_5, rest0_4, v_7, w, yaml_ast_list_3) = (None, None, None, None, None)
                                if not is_empty(restlist):
                                    active_pattern_result_11: dict[str, Any] | None = _007CKey_007C__007C(head(restlist))
                                    if active_pattern_result_11 is not None:
                                        if not is_empty(tail(restlist)):
                                            active_pattern_result_12: dict[str, Any] | None = _007CSequenceMinusOpener_007C__007C(head(tail(restlist)))
                                            if active_pattern_result_12 is not None:
                                                if not is_empty(tail(tail(restlist))):
                                                    if head(tail(tail(restlist))).tag == 1:
                                                        pattern_matching_result_5 = 0
                                                        rest0_4 = tail(tail(tail(restlist)))
                                                        v_7 = active_pattern_result_11
                                                        w = active_pattern_result_12
                                                        yaml_ast_list_3 = head(tail(tail(restlist))).fields[0]

                                                    else: 
                                                        pattern_matching_result_5 = 1


                                                else: 
                                                    pattern_matching_result_5 = 1


                                            else: 
                                                pattern_matching_result_5 = 1


                                        else: 
                                            pattern_matching_result_5 = 1


                                    else: 
                                        pattern_matching_result_5 = 1


                                else: 
                                    pattern_matching_result_5 = 1

                                if pattern_matching_result_5 == 0:
                                    c_3: str | None = restore_comment_replace(comment_dict, v_7["Comment"])
                                    object_list_2: FSharpList[PreprocessorElement] = cons(PreprocessorElement(2, value_2(w["Value"])), yaml_ast_list_3) if (w["Value"] is not None) else yaml_ast_list_3
                                    def _arrow461(e_8: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                                        return is_sequence_element(e_8)

                                    sequence_elements_4: FSharpList[FSharpList[PreprocessorElement]] = collect_sequence_elements(to_list(take_while(_arrow461, rest0_4)))
                                    def _arrow462(e_9: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                                        return is_sequence_element(e_9)

                                    rest_9: FSharpList[PreprocessorElement] = to_list(skip_while(_arrow462, rest0_4))
                                    def _arrow465(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                                        def _arrow464(__unit: None=None) -> IEnumerable_1[YAMLElement]:
                                            def _arrow463(i_6: FSharpList[PreprocessorElement]) -> YAMLElement:
                                                return loop_read(i_6, empty())

                                            return map_1(_arrow463, sequence_elements_4)

                                        return append(singleton(loop_read(object_list_2, empty())), delay(_arrow464))

                                    seq: YAMLElement = YAMLElement(2, to_list(delay(_arrow465)))
                                    restlist_mut = rest_9
                                    acc_mut = cons(YAMLElement(0, YAMLContent_create_27AED5E3(v_7["Key"], c_3), YAMLElement(3, singleton_1(seq))), acc)
                                    continue

                                elif pattern_matching_result_5 == 1:
                                    (pattern_matching_result_6, rest0_5, v_8, w_1) = (None, None, None, None)
                                    if not is_empty(restlist):
                                        active_pattern_result_13: dict[str, Any] | None = _007CKey_007C__007C(head(restlist))
                                        if active_pattern_result_13 is not None:
                                            if not is_empty(tail(restlist)):
                                                active_pattern_result_14: dict[str, Any] | None = _007CSequenceMinusOpener_007C__007C(head(tail(restlist)))
                                                if active_pattern_result_14 is not None:
                                                    pattern_matching_result_6 = 0
                                                    rest0_5 = tail(tail(restlist))
                                                    v_8 = active_pattern_result_13
                                                    w_1 = active_pattern_result_14

                                                else: 
                                                    pattern_matching_result_6 = 1


                                            else: 
                                                pattern_matching_result_6 = 1


                                        else: 
                                            pattern_matching_result_6 = 1


                                    else: 
                                        pattern_matching_result_6 = 1

                                    if pattern_matching_result_6 == 0:
                                        c_4: str | None = restore_comment_replace(comment_dict, v_8["Comment"])
                                        def _arrow466(e_10: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                                            return is_sequence_element(e_10)

                                        sequence_elements_5: FSharpList[FSharpList[PreprocessorElement]] = collect_sequence_elements(to_list(take_while(_arrow466, rest0_5)))
                                        def _arrow467(e_11: PreprocessorElement, restlist: Any=restlist, acc: Any=acc) -> bool:
                                            return is_sequence_element(e_11)

                                        rest_10: FSharpList[PreprocessorElement] = to_list(skip_while(_arrow467, rest0_5))
                                        def _arrow470(__unit: None=None, restlist: Any=restlist, acc: Any=acc) -> IEnumerable_1[YAMLElement]:
                                            def _arrow469(__unit: None=None) -> IEnumerable_1[YAMLElement]:
                                                def _arrow468(i_7: FSharpList[PreprocessorElement]) -> YAMLElement:
                                                    return loop_read(i_7, empty())

                                                return map_1(_arrow468, sequence_elements_5)

                                            return append(singleton(loop_read(singleton_1(PreprocessorElement(2, value_2(w_1["Value"]))), empty())), delay(_arrow469))

                                        seq_1: YAMLElement = YAMLElement(2, to_list(delay(_arrow470)))
                                        restlist_mut = rest_10
                                        acc_mut = cons(YAMLElement(0, YAMLContent_create_27AED5E3(v_8["Key"], c_4), YAMLElement(3, singleton_1(seq_1))), acc)
                                        continue

                                    elif pattern_matching_result_6 == 1:
                                        if is_empty(restlist):
                                            return YAMLElement(3, reverse(acc))

                                        else: 
                                            active_pattern_result_15: dict[str, Any] | None = _007CKeyValue_007C__007C(head(restlist))
                                            if active_pattern_result_15 is not None:
                                                v_9: dict[str, Any] = active_pattern_result_15
                                                restlist_mut = tail(restlist)
                                                acc_mut = cons(YAMLElement(0, YAMLContent_create_27AED5E3(v_9["Key"]), loop_read(singleton_1(PreprocessorElement(2, v_9["Value"])), empty())), acc)
                                                continue

                                            else: 
                                                active_pattern_result_16: dict[str, Any] | None = _007CYamlComment_007C__007C(head(restlist))
                                                if active_pattern_result_16 is not None:
                                                    v_10: dict[str, Any] = active_pattern_result_16
                                                    restlist_mut = tail(restlist)
                                                    acc_mut = cons(YAMLElement(4, get_item_from_dict(comment_dict, v_10["Comment"])), acc)
                                                    continue

                                                else: 
                                                    active_pattern_result_17: dict[str, Any] | None = _007CYamlValue_007C__007C(head(restlist))
                                                    if active_pattern_result_17 is not None:
                                                        v_11: dict[str, Any] = active_pattern_result_17
                                                        restlist_mut = tail(restlist)
                                                        acc_mut = cons(YAMLElement(1, YAMLContent_create_27AED5E3(restore_string_replace(string_dict, v_11["Value"]), restore_comment_replace(comment_dict, v_11["Comment"]))), acc)
                                                        continue

                                                    else: 
                                                        return to_fail(printf("Unknown pattern: %A"))(restlist)











            break

    return loop_read(yaml_list, empty())


def read(yaml: str) -> YAMLElement:
    ast: Preprocessor = read_1(yaml)
    match_value: PreprocessorElement = ast.AST
    if match_value.tag == 0:
        return tokenize(match_value.fields[0], ast.StringMap, ast.CommentMap)

    else: 
        raise Exception("Not a root!")



__all__ = ["restore_string_replace", "restore_comment_replace", "collect_sequence_elements", "is_sequence_element", "tokenize", "read"]




from __future__ import annotations
from datetime import datetime
from decimal import Decimal
from typing import Any
from ...fable_library.char import parse as parse_6
from ...fable_library.date import parse as parse_4
from ...fable_library.decimal_ import parse as parse_3
from ...fable_library.double import parse
from ...fable_library.guid import parse as parse_5
from ...fable_library.int32 import parse as parse_1
from ...fable_library.long import parse as parse_2
from ...fable_library.option import (default_arg, map)
from ...fable_library.reflection import (TypeInfo, union_type, class_type)
from ...fable_library.seq import for_all
from ...fable_library.types import (Array, Union, to_string, uint32, int64, uint64)
from ...fable_library.util import (equals, to_enumerable)
from ..fs_address import (FsAddress, CellReference_indexToColAdress)

def _expr107() -> TypeInfo:
    return union_type("FsSpreadsheet.DataType", [], DataType, lambda: [[], [], [], [], []])


class DataType(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["String", "Boolean", "Number", "Date", "Empty"]


DataType_reflection = _expr107

def FsCellAux_boolConverter(bool_1: bool) -> str:
    if bool_1:
        return "1"

    else: 
        return "0"



def _expr108() -> TypeInfo:
    return class_type("FsSpreadsheet.FsCell", None, FsCell)


class FsCell:
    def __init__(self, value: Any=None, data_type: DataType | None=None, address: FsAddress | None=None) -> None:
        self._cellValue: Any = value
        self._dataType: DataType = default_arg(data_type, DataType(0))
        self._comment: str = ""
        self._hyperlink: str = ""
        self._richText: str = ""
        self._formulaA1: str = ""
        self._formulaR1C1: str = ""
        def mapping(a: FsAddress) -> int:
            return a.RowNumber

        self._rowIndex: int = default_arg(map(mapping, address), 1) or 0
        def mapping_1(a_1: FsAddress) -> int:
            return a_1.ColumnNumber

        self._columnIndex: int = default_arg(map(mapping_1, address), 1) or 0

    @property
    def Value(self, __unit: None=None) -> Any:
        self_1: FsCell = self
        return self_1._cellValue

    @Value.setter
    def Value(self, value: Any=None) -> None:
        self_1: FsCell = self
        self_1._cellValue = value

    @property
    def DataType(self, __unit: None=None) -> DataType:
        self_1: FsCell = self
        return self_1._dataType

    @DataType.setter
    def DataType(self, data_type: DataType) -> None:
        self_1: FsCell = self
        self_1._dataType = data_type

    @property
    def ColumnNumber(self, __unit: None=None) -> int:
        self_1: FsCell = self
        return self_1._columnIndex

    @ColumnNumber.setter
    def ColumnNumber(self, col_i: int) -> None:
        self_1: FsCell = self
        self_1._columnIndex = col_i or 0

    @property
    def RowNumber(self, __unit: None=None) -> int:
        self_1: FsCell = self
        return self_1._rowIndex

    @RowNumber.setter
    def RowNumber(self, row_i: int) -> None:
        self_1: FsCell = self
        self_1._rowIndex = row_i or 0

    @property
    def Address(self, __unit: None=None) -> FsAddress:
        self_1: FsCell = self
        return FsAddress(self_1._rowIndex, self_1._columnIndex)

    @Address.setter
    def Address(self, address: FsAddress) -> None:
        self_1: FsCell = self
        self_1._rowIndex = address.RowNumber or 0
        self_1._columnIndex = address.ColumnNumber or 0

    @staticmethod
    def create(row_number: int, col_number: int, value: Any) -> FsCell:
        pattern_input: tuple[DataType, Any]
        value_1_1: Any = value
        pattern_input = ((DataType(0), value_1_1)) if (str(type(value_1_1)) == "<class \'str\'>") else ((((DataType(1), True)) if value_1_1 else ((DataType(1), False))) if (str(type(value_1_1)) == "<class \'bool\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<fable_modules.fable_library.types.uint8\'>>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int8\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'int\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int16\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int64\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint32>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint16\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint32\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.float32\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'float\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'int\'>") else (((DataType(3), value_1_1)) if isinstance(value_1_1, datetime) else (((DataType(0), value_1_1)) if (str(type(value_1_1)) == "<class \'str\'>") else ((DataType(0), value_1_1))))))))))))))))
        return FsCell(pattern_input[1], pattern_input[0], FsAddress(row_number, col_number))

    @staticmethod
    def create_empty(__unit: None=None) -> FsCell:
        return FsCell("", DataType(4), FsAddress(0, 0))

    @staticmethod
    def create_with_adress(adress: FsAddress, value: Any) -> FsCell:
        pattern_input: tuple[DataType, Any]
        value_1_1: Any = value
        pattern_input = ((DataType(0), value_1_1)) if (str(type(value_1_1)) == "<class \'str\'>") else ((((DataType(1), True)) if value_1_1 else ((DataType(1), False))) if (str(type(value_1_1)) == "<class \'bool\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<fable_modules.fable_library.types.uint8\'>>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int8\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'int\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int16\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int64\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint32>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint16\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint32\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.float32\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'float\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'int\'>") else (((DataType(3), value_1_1)) if isinstance(value_1_1, datetime) else (((DataType(0), value_1_1)) if (str(type(value_1_1)) == "<class \'str\'>") else ((DataType(0), value_1_1))))))))))))))))
        return FsCell(pattern_input[1], pattern_input[0], adress)

    @staticmethod
    def create_empty_with_adress(adress: FsAddress) -> FsCell:
        return FsCell("", DataType(4), adress)

    @staticmethod
    def create_with_data_type(data_type: DataType, row_number: int, col_number: int, value: Any=None) -> FsCell:
        return FsCell(value, data_type, FsAddress(row_number, col_number))

    def __str__(self, __unit: None=None) -> str:
        self_1: FsCell = self
        return ((((((("" + CellReference_indexToColAdress(int(self_1.ColumnNumber+0x100000000 if self_1.ColumnNumber < 0 else self_1.ColumnNumber))) + "") + str(self_1.RowNumber)) + " : ") + str(self_1.Value)) + " | ") + str(self_1.DataType)) + ""

    def CopyFrom(self, other_cell: FsCell) -> None:
        self_1: FsCell = self
        self_1.DataType = other_cell.DataType
        self_1.Value = other_cell.Value

    def CopyTo(self, target: FsCell) -> None:
        self_1: FsCell = self
        target.DataType = self_1.DataType
        target.Value = self_1.Value

    @staticmethod
    def copy_from_to(source_cell: FsCell, target_cell: FsCell) -> FsCell:
        target_cell.DataType = source_cell.DataType
        target_cell.Value = source_cell.Value
        return target_cell

    def Copy(self, __unit: None=None) -> FsCell:
        self_1: FsCell = self
        return FsCell(self_1.Value, self_1.DataType, self_1.Address.Copy())

    @staticmethod
    def copy(cell: FsCell) -> FsCell:
        return cell.Copy()

    def ValueAsString(self, __unit: None=None) -> str:
        self_1: FsCell = self
        v: Any = self_1.Value
        match_value: DataType = self_1.DataType
        if ((match_value.tag == 3) or (match_value.tag == 1)) or (match_value.tag == 4):
            return to_string(v)

        elif match_value.tag == 2:
            return to_string(v)

        else: 
            return to_string(v)


    @staticmethod
    def get_value_as_string(cell: FsCell) -> str:
        return cell.ValueAsString()

    def ValueAsBool(self, __unit: None=None) -> bool:
        self_1: FsCell = self
        match_value: str = to_string(self_1.Value).lower()
        (pattern_matching_result,) = (None,)
        if match_value == "1":
            pattern_matching_result = 0

        elif match_value == "true":
            pattern_matching_result = 0

        elif match_value == "true()":
            pattern_matching_result = 0

        elif match_value == "0":
            pattern_matching_result = 1

        elif match_value == "false":
            pattern_matching_result = 1

        elif match_value == "false()":
            pattern_matching_result = 1

        else: 
            pattern_matching_result = 2

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return False

        elif pattern_matching_result == 2:
            raise Exception(("String \'" + match_value) + "\' was not recognized as a valid Boolean")


    @staticmethod
    def get_value_as_bool(cell: FsCell) -> bool:
        return cell.ValueAsBool()

    def ValueAsFloat(self, __unit: None=None) -> float:
        self_1: FsCell = self
        return parse(to_string(self_1.Value))

    @staticmethod
    def get_value_as_float(cell: FsCell) -> float:
        return cell.ValueAsFloat()

    def ValueAsInt(self, __unit: None=None) -> int:
        self_1: FsCell = self
        return parse_1(to_string(self_1.Value), 511, False, 32)

    @staticmethod
    def get_value_as_int(cell: FsCell) -> int:
        return cell.ValueAsInt()

    def ValueAsUInt(self, __unit: None=None) -> uint32:
        self_1: FsCell = self
        return parse_1(to_string(self_1.Value), 511, True, 32)

    @staticmethod
    def get_value_as_uint(cell: FsCell) -> uint32:
        return cell.ValueAsUInt()

    def ValueAsLong(self, __unit: None=None) -> int64:
        self_1: FsCell = self
        return parse_2(to_string(self_1.Value), 511, False, 64)

    @staticmethod
    def get_value_as_long(cell: FsCell) -> int64:
        return cell.ValueAsLong()

    def ValueAsULong(self, __unit: None=None) -> uint64:
        self_1: FsCell = self
        return parse_2(to_string(self_1.Value), 511, True, 64)

    @staticmethod
    def get_value_as_ulong(cell: FsCell) -> uint64:
        return cell.ValueAsULong()

    def ValueAsDouble(self, __unit: None=None) -> float:
        self_1: FsCell = self
        return parse(to_string(self_1.Value))

    @staticmethod
    def get_value_as_double(cell: FsCell) -> float:
        return cell.ValueAsDouble()

    def ValueAsDecimal(self, __unit: None=None) -> Decimal:
        self_1: FsCell = self
        return parse_3(to_string(self_1.Value))

    @staticmethod
    def get_value_as_decimal(cell: FsCell) -> Decimal:
        return cell.ValueAsDecimal()

    def ValueAsDateTime(self, __unit: None=None) -> Any:
        self_1: FsCell = self
        return parse_4(to_string(self_1.Value))

    @staticmethod
    def get_value_as_date_time(cell: FsCell) -> Any:
        return cell.ValueAsDateTime()

    def ValueAsGuid(self, __unit: None=None) -> str:
        self_1: FsCell = self
        return parse_5(to_string(self_1.Value))

    @staticmethod
    def get_value_as_guid(cell: FsCell) -> str:
        return cell.ValueAsGuid()

    def ValueAsChar(self, __unit: None=None) -> str:
        self_1: FsCell = self
        return parse_6(to_string(self_1.Value))

    @staticmethod
    def get_value_as_char(cell: FsCell) -> str:
        return cell.ValueAsChar()

    def SetValueAs(self, value: Any=None) -> None:
        self_1: FsCell = self
        pattern_input: tuple[DataType, Any]
        value_1_1: Any = value
        pattern_input = ((DataType(0), value_1_1)) if (str(type(value_1_1)) == "<class \'str\'>") else ((((DataType(1), True)) if value_1_1 else ((DataType(1), False))) if (str(type(value_1_1)) == "<class \'bool\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<fable_modules.fable_library.types.uint8\'>>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int8\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'int\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int16\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.int64\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint32>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint16\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.uint32\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'fable_modules.fable_library.types.float32\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'float\'>") else (((DataType(2), value_1_1)) if (str(type(value_1_1)) == "<class \'int\'>") else (((DataType(3), value_1_1)) if isinstance(value_1_1, datetime) else (((DataType(0), value_1_1)) if (str(type(value_1_1)) == "<class \'str\'>") else ((DataType(0), value_1_1))))))))))))))))
        self_1._dataType = pattern_input[0]
        self_1._cellValue = pattern_input[1]

    @staticmethod
    def set_value_as(value: Any, cell: FsCell) -> FsCell:
        cell.SetValueAs(value)
        return cell

    def StructurallyEquals(self, other: FsCell) -> bool:
        this: FsCell = self
        def predicate_1(x_1: bool) -> bool:
            return x_1 == True

        def predicate(x: bool) -> bool:
            return x == True

        return for_all(predicate_1, to_enumerable([equals(this.Value, other.Value), equals(this.DataType, other.DataType), for_all(predicate, to_enumerable([this.Address.Address == other.Address.Address, this.Address.ColumnNumber == other.Address.ColumnNumber, this.Address.RowNumber == other.Address.RowNumber, this.Address.FixedColumn == other.Address.FixedColumn, this.Address.FixedRow == other.Address.FixedRow])), this.ColumnNumber == other.ColumnNumber, this.RowNumber == other.RowNumber]))


FsCell_reflection = _expr108

def FsCell__ctor_2BEF9BB0(value: Any=None, data_type: DataType | None=None, address: FsAddress | None=None) -> FsCell:
    return FsCell(value, data_type, address)


__all__ = ["DataType_reflection", "FsCellAux_boolConverter", "FsCell_reflection"]




from __future__ import annotations
from datetime import datetime
from decimal import Decimal
from typing import (Any, TypeVar)
from ...arctrl_rocrate.ldobject import LDValue
from ...fable_library.list import (of_array, FSharpList)
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (one_of, map, string, int_1, decimal, object, IRequiredGetter, IOptionalGetter, IGetters)
from ...thoth_json_core.types import (Decoder_1, IEncodable, IEncoderHelpers_1)
from ..encode import date_time

__A_ = TypeVar("__A_")

def _arrow1887(value: str) -> Any:
    return value


def _arrow1888(value_1: int) -> Any:
    return value_1


def _arrow1889(value_2: Decimal) -> Any:
    return value_2


generic_decoder: Decoder_1[Any] = one_of(of_array([map(_arrow1887, string), map(_arrow1888, int_1), map(_arrow1889, decimal)]))

def generic_encoder(value: Any=None) -> IEncodable:
    if str(type(value)) == "<class \'str\'>":
        class ObjectExpr1890(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr1890()

    elif str(type(value)) == "<class \'int\'>":
        class ObjectExpr1891(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_1.encode_signed_integral_number(value)

        return ObjectExpr1891()

    elif str(type(value)) == "<class \'bool\'>":
        class ObjectExpr1892(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_2.encode_bool(value)

        return ObjectExpr1892()

    elif str(type(value)) == "<class \'float\'>":
        class ObjectExpr1893(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_3.encode_decimal_number(value)

        return ObjectExpr1893()

    elif isinstance(value, datetime):
        return date_time(value)

    else: 
        raise Exception("Unknown type")



def _arrow1896(decoders: IGetters) -> LDValue:
    def _arrow1894(__unit: None=None) -> Any:
        object_arg: IRequiredGetter = decoders.Required
        return object_arg.Field("@value", generic_decoder)

    def _arrow1895(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = decoders.Optional
        return object_arg_1.Field("@type", string)

    return LDValue(_arrow1894(), _arrow1895())


decoder: Decoder_1[LDValue] = object(_arrow1896)

def encoder(v: LDValue) -> IEncodable:
    def _arrow1898(__unit: None=None, v: Any=v) -> IEncodable:
        value: str = v.ValueType
        class ObjectExpr1897(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr1897()

    values: FSharpList[tuple[str, IEncodable]] = of_array([("@value", generic_encoder(v.Value)), ("@type", _arrow1898())])
    class ObjectExpr1899(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], v: Any=v) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr1899()


__all__ = ["generic_decoder", "generic_encoder", "decoder", "encoder"]

